#
# Copyright (c) 2018 Sorin Sbarnea <ssbarnea@users.noreply.github.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tests import base
from tests.base import mock
import os
from jenkins_jobs.modules import project_multibranch


@mock.patch("uuid.uuid4", mock.Mock(return_value="1-1-1-1-1"))
class TestCaseMultibranchPipeline(base.BaseScenariosTestCase):
    fixtures_path = os.path.join(os.path.dirname(__file__), "fixtures")
    scenarios = base.get_scenarios(fixtures_path)
    default_config_file = "/dev/null"
    klass = project_multibranch.WorkflowMultiBranch



#!/usr/bin/env python
# Copyright (C) 2015 OpenStack, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Manage JJB yaml feature implementation

import copy
import fnmatch
import io
import itertools
import logging
import re
import os

from jenkins_jobs.constants import MAGIC_MANAGE_STRING
from jenkins_jobs.errors import JenkinsJobsException
from jenkins_jobs.formatter import deep_format
import jenkins_jobs.local_yaml as local_yaml
from jenkins_jobs import utils

__all__ = ["YamlParser"]

logger = logging.getLogger(__name__)


def matches(what, glob_patterns):
    """
    Checks if the given string, ``what``, matches any of the glob patterns in
    the iterable, ``glob_patterns``

    :arg str what: String that we want to test if it matches a pattern
    :arg iterable glob_patterns: glob patterns to match (list, tuple, set,
    etc.)
    """
    return any(fnmatch.fnmatch(what, glob_pattern) for glob_pattern in glob_patterns)


def combination_matches(combination, match_combinations):
    """
    Checks if the given combination is matches for any of the given combination
    globs, being those a set of combinations where if a key is missing, it's
    considered matching

    (key1=2, key2=3)

    would match the combination match:
    (key2=3)

    but not:
    (key1=2, key2=2)
    """
    for cmatch in match_combinations:
        for key, val in combination.items():
            if cmatch.get(key, val) != val:
                break
        else:
            return True
    return False


class YamlParser(object):
    def __init__(self, jjb_config=None):
        self.data = {}
        self.jobs = []
        self.views = []

        self.jjb_config = jjb_config
        self.keep_desc = jjb_config.yamlparser["keep_descriptions"]
        self.path = jjb_config.yamlparser["include_path"]

    def load_files(self, fn):

        # handle deprecated behavior, and check that it's not a file like
        # object as these may implement the '__iter__' attribute.
        if not hasattr(fn, "__iter__") or hasattr(fn, "read"):
            logger.warning(
                "Passing single elements for the `fn` argument in "
                "Builder.load_files is deprecated. Please update your code "
                "to use a list as support for automatic conversion will be "
                "removed in a future version."
            )
            fn = [fn]

        files_to_process = []
        for path in fn:
            if not hasattr(path, "read") and os.path.isdir(path):
                files_to_process.extend(
                    [
                        os.path.join(path, f)
                        for f in sorted(os.listdir(path))
                        if (f.endswith(".yml") or f.endswith(".yaml"))
                    ]
                )
            else:
                files_to_process.append(path)

        # symlinks used to allow loading of sub-dirs can result in duplicate
        # definitions of macros and templates when loading all from top-level
        unique_files = []
        for f in files_to_process:
            if hasattr(f, "read"):
                unique_files.append(f)
                continue
            rpf = os.path.realpath(f)
            if rpf not in unique_files:
                unique_files.append(rpf)
            else:
                logger.warning(
                    "File '%s' already added as '%s', ignoring "
                    "reference to avoid duplicating yaml "
                    "definitions." % (f, rpf)
                )

        for in_file in unique_files:
            # use of ask-for-permissions instead of ask-for-forgiveness
            # performs better when low use cases.
            if hasattr(in_file, "name"):
                fname = in_file.name
            else:
                fname = in_file
            logger.debug("Parsing YAML file {0}".format(fname))
            if hasattr(in_file, "read"):
                self._parse_fp(in_file)
            else:
                self.parse(in_file)

    def _parse_fp(self, fp):
        # wrap provided file streams to ensure correct encoding used
        data = local_yaml.load(
            utils.wrap_stream(fp),
            self.jjb_config.yamlparser["retain_anchors"],
            search_path=self.path,
        )
        if data:
            if not isinstance(data, list):
                raise JenkinsJobsException(
                    "The topmost collection in file '{fname}' must be a list,"
                    " not a {cls}".format(fname=getattr(fp, "name", fp), cls=type(data))
                )
            for item in data:
                cls, dfn = next(iter(item.items()))
                group = self.data.get(cls, {})
                if len(item.items()) > 1:
                    n = None
                    for k, v in item.items():
                        if k == "name":
                            n = v
                            break
                    # Syntax error
                    raise JenkinsJobsException(
                        "Syntax error, for item "
                        "named '{0}'. Missing indent?".format(n)
                    )
                # allow any entry to specify an id that can also be used
                _id = dfn.get("id", dfn["name"])
                if _id in group:
                    self._handle_dups(
                        "Duplicate entry found in '{0}: '{1}' already "
                        "defined".format(fp.name, _id)
                    )
                group[_id] = dfn
                self.data[cls] = group

    def parse(self, fn):
        with io.open(fn, "r", encoding="utf-8") as fp:
            self._parse_fp(fp)

    def _handle_dups(self, message):

        if not self.jjb_config.yamlparser["allow_duplicates"]:
            logger.error(message)
            raise JenkinsJobsException(message)
        else:
            logger.warning(message)

    def _getJob(self, name):
        job = self.data.get("job", {}).get(name, None)
        if not job:
            return job
        return self._applyDefaults(job)

    def _getJobGroup(self, name):
        return self.data.get("job-group", {}).get(name, None)

    def _getJobTemplate(self, name):
        job = self.data.get("job-template", {}).get(name, None)
        if not job:
            return job
        return self._applyDefaults(job)

    def _applyDefaults(self, data, override_dict=None):
        if override_dict is None:
            override_dict = {}

        whichdefaults = data.get("defaults", "global")
        defaults = copy.deepcopy(self.data.get("defaults", {}).get(whichdefaults, {}))
        if defaults == {} and whichdefaults != "global":
            raise JenkinsJobsException(
                "Unknown defaults set: '{0}'".format(whichdefaults)
            )

        for key in override_dict.keys():
            if key in defaults.keys():
                defaults[key] = override_dict[key]

        newdata = {}
        newdata.update(defaults)
        newdata.update(data)
        return newdata

    def _formatDescription(self, job):
        if self.keep_desc:
            description = job.get("description", None)
        else:
            description = job.get("description", "")
        if description is not None:
            job["description"] = description + self._get_managed_string().lstrip()

    def _getfullname(self, data):
        if "folder" in data:
            return "%s/%s" % (data["folder"], data["name"])

        return data["name"]

    def expandYaml(self, registry, jobs_glob=None):
        changed = True
        while changed:
            changed = False
            for module in registry.modules:
                if hasattr(module, "handle_data"):
                    if module.handle_data(self.data):
                        changed = True

        for job in self.data.get("job", {}).values():
            job = self._applyDefaults(job)
            job["name"] = self._getfullname(job)

            if jobs_glob and not matches(job["name"], jobs_glob):
                logger.debug("Ignoring job {0}".format(job["name"]))
                continue
            logger.debug("Expanding job '{0}'".format(job["name"]))
            self._formatDescription(job)
            self.jobs.append(job)

        for view in self.data.get("view", {}).values():
            view["name"] = self._getfullname(view)

            if jobs_glob and not matches(view["name"], jobs_glob):
                logger.debug("Ignoring view {0}".format(view["name"]))
                continue

            logger.debug("Expanding view '{0}'".format(view["name"]))
            self._formatDescription(view)
            self.views.append(view)

        for project in self.data.get("project", {}).values():
            logger.debug("Expanding project '{0}'".format(project["name"]))
            # use a set to check for duplicate job references in projects
            seen = set()
            for jobspec in project.get("jobs", []):
                if isinstance(jobspec, dict):
                    # Singleton dict containing dict of job-specific params
                    jobname, jobparams = next(iter(jobspec.items()))
                    if not isinstance(jobparams, dict):
                        jobparams = {}
                else:
                    jobname = jobspec
                    jobparams = {}
                job = self._getJob(jobname)
                if job:
                    # Just naming an existing defined job
                    if jobname in seen:
                        self._handle_dups(
                            "Duplicate job '{0}' specified "
                            "for project '{1}'".format(jobname, project["name"])
                        )
                    seen.add(jobname)
                    continue
                # see if it's a job group
                group = self._getJobGroup(jobname)
                if group:
                    for group_jobspec in group["jobs"]:
                        if isinstance(group_jobspec, dict):
                            group_jobname, group_jobparams = next(
                                iter(group_jobspec.items())
                            )
                            if not isinstance(group_jobparams, dict):
                                group_jobparams = {}
                        else:
                            group_jobname = group_jobspec
                            group_jobparams = {}
                        job = self._getJob(group_jobname)
                        if job:
                            if group_jobname in seen:
                                self._handle_dups(
                                    "Duplicate job '{0}' specified for "
                                    "project '{1}'".format(
                                        group_jobname, project["name"]
                                    )
                                )
                            seen.add(group_jobname)
                            continue
                        template = self._getJobTemplate(group_jobname)
                        # Allow a group to override parameters set by a project
                        d = type(project)(project)
                        d.update(jobparams)
                        d.update(group)
                        d.update(group_jobparams)
                        # Except name, since the group's name is not useful
                        d["name"] = project["name"]
                        if template:
                            self._expandYamlForTemplateJob(d, template, jobs_glob)
                    continue
                # see if it's a template
                template = self._getJobTemplate(jobname)
                if template:
                    d = type(project)(project)
                    d.update(jobparams)
                    self._expandYamlForTemplateJob(d, template, jobs_glob)
                else:
                    raise JenkinsJobsException(
                        "Failed to find suitable "
                        "template named '{0}'".format(jobname)
                    )

            for viewspec in project.get("views", []):
                if isinstance(viewspec, dict):
                    # Singleton dict containing dict of view-specific params
                    viewname, viewparams = next(iter(viewspec.items()))
                    if not isinstance(viewparams, dict):
                        viewparams = {}
                else:
                    viewname = viewspec
                    viewparams = {}
                view = self._getView(viewname)
                if view:
                    # Just naming an existing defined view
                    if viewname in seen:
                        self._handle_dups(
                            "Duplicate view '{0}' specified "
                            "for project '{1}'".format(viewname, project["name"])
                        )
                    seen.add(viewname)
                    continue
                # see if it's a view group
                group = self._getViewGroup(viewname)
                if group:
                    for group_viewspec in group["views"]:
                        if isinstance(group_viewspec, dict):
                            group_viewname, group_viewparams = next(
                                iter(group_viewspec.items())
                            )
                            if not isinstance(group_viewparams, dict):
                                group_viewparams = {}
                        else:
                            group_viewname = group_viewspec
                            group_viewparams = {}
                        view = self._getView(group_viewname)
                        if view:
                            if group_viewname in seen:
                                self._handle_dups(
                                    "Duplicate view '{0}' specified for "
                                    "project '{1}'".format(
                                        group_viewname, project["name"]
                                    )
                                )
                            seen.add(group_viewname)
                            continue
                        template = self._getViewTemplate(group_viewname)
                        # Allow a group to override parameters set by a project
                        d = type(project)(project)
                        d.update(viewparams)
                        d.update(group)
                        d.update(group_viewparams)
                        # Except name, since the group's name is not useful
                        d["name"] = project["name"]
                        if template:
                            self._expandYamlForTemplateView(d, template, jobs_glob)
                    continue
                # see if it's a template
                template = self._getViewTemplate(viewname)
                if template:
                    d = type(project)(project)
                    d.update(viewparams)
                    self._expandYamlForTemplateView(d, template, jobs_glob)
                else:
                    raise JenkinsJobsException(
                        "Failed to find suitable "
                        "template named '{0}'".format(viewname)
                    )

        # check for duplicate generated jobs
        seen = set()
        # walk the list in reverse so that last definition wins
        for job in self.jobs[::-1]:
            if job["name"] in seen:
                self._handle_dups(
                    "Duplicate definitions for job '{0}' "
                    "specified".format(job["name"])
                )
                self.jobs.remove(job)
            seen.add(job["name"])

        # check for duplicate generated views
        seen_views = set()
        # walk the list in reverse so that last definition wins
        for view in self.views[::-1]:
            if view["name"] in seen_views:
                self._handle_dups(
                    "Duplicate definitions for view '{0}' "
                    "specified".format(view["name"])
                )
                self.views.remove(view)
            seen_views.add(view["name"])

        return self.jobs, self.views

    def _expandYamlForTemplateJob(self, project, template, jobs_glob=None):
        dimensions = []
        template_name = template["name"]
        # reject keys that are not useful during yaml expansion
        for k in ["jobs"]:
            project.pop(k)
        excludes = project.pop("exclude", [])
        for (k, v) in project.items():
            tmpk = "{{{0}}}".format(k)
            if tmpk not in template_name:
                continue
            if type(v) == list:
                dimensions.append(zip([k] * len(v), v))
        # XXX somewhat hackish to ensure we actually have a single
        # pass through the loop
        if len(dimensions) == 0:
            dimensions = [(("", ""),)]

        for values in itertools.product(*dimensions):
            params = copy.deepcopy(project)
            params = self._applyDefaults(params, template)
            params["template-name"] = re.sub(r"({|})", r"\1\1", template_name)

            try:
                expanded_values = {}
                for (k, v) in values:
                    if isinstance(v, dict):
                        inner_key = next(iter(v))
                        expanded_values[k] = inner_key
                        expanded_values.update(v[inner_key])
                    else:
                        expanded_values[k] = v
            except TypeError:
                project_name = project.pop("name")
                logger.error(
                    "Exception thrown while expanding template '%s' for "
                    "project '%s', with expansion arguments of:\n%s\n"
                    "Original project input variables for template:\n%s\n"
                    "Most likely the inputs have items indented incorrectly "
                    "to describe how they should be applied.\n\nNote yaml "
                    "'null' is mapped to python's 'None'",
                    template_name,
                    project_name,
                    "".join(
                        local_yaml.dump({k: v}, default_flow_style=False)
                        for (k, v) in values
                    ),
                    local_yaml.dump(project, default_flow_style=False),
                )
                raise

            params.update(expanded_values)
            try:
                params = deep_format(params, params)
            except Exception:
                logging.error("Failure formatting params '%s' with itself", params)
                raise
            if combination_matches(params, excludes):
                logger.debug("Excluding combination %s", str(params))
                continue

            for key in template.keys():
                if key not in params:
                    params[key] = template[key]

            try:
                expanded = deep_format(
                    template,
                    params,
                    self.jjb_config.yamlparser["allow_empty_variables"],
                )
            except Exception:
                logging.error(
                    "Failure formatting template '%s', containing '%s' with "
                    "params '%s'",
                    template_name,
                    template,
                    params,
                )
                raise
            expanded["name"] = self._getfullname(expanded)

            job_name = expanded.get("name")
            if jobs_glob and not matches(job_name, jobs_glob):
                continue

            self._formatDescription(expanded)
            self.jobs.append(expanded)

    def _get_managed_string(self):
        # The \n\n is not hard coded, because they get stripped if the
        # project does not otherwise have a description.
        return "\n\n" + MAGIC_MANAGE_STRING

    # Views related
    def _getView(self, name):
        view = self.data.get("view", {}).get(name, None)
        if not view:
            return view
        return self._applyDefaults(view)

    def _getViewGroup(self, name):
        return self.data.get("view-group", {}).get(name, None)

    def _getViewTemplate(self, name):
        view = self.data.get("view-template", {}).get(name, None)
        if not view:
            return view
        return self._applyDefaults(view)

    def _expandYamlForTemplateView(self, project, template, views_glob=None):
        dimensions = []
        template_name = template["name"]
        # reject keys that are not useful during yaml expansion
        for k in ["views"]:
            project.pop(k)
        excludes = project.pop("exclude", [])
        for (k, v) in project.items():
            tmpk = "{{{0}}}".format(k)
            if tmpk not in template_name:
                continue
            if type(v) == list:
                dimensions.append(zip([k] * len(v), v))
        # XXX somewhat hackish to ensure we actually have a single
        # pass through the loop
        if len(dimensions) == 0:
            dimensions = [(("", ""),)]

        for values in itertools.product(*dimensions):
            params = copy.deepcopy(project)
            params = self._applyDefaults(params, template)

            expanded_values = {}
            for (k, v) in values:
                if isinstance(v, dict):
                    inner_key = next(iter(v))
                    expanded_values[k] = inner_key
                    expanded_values.update(v[inner_key])
                else:
                    expanded_values[k] = v

            params.update(expanded_values)
            params = deep_format(params, params)
            if combination_matches(params, excludes):
                logger.debug("Excluding combination %s", str(params))
                continue

            for key in template.keys():
                if key not in params:
                    params[key] = template[key]

            params["template-name"] = template_name
            expanded = deep_format(
                template, params, self.jjb_config.yamlparser["allow_empty_variables"]
            )

            view_name = expanded.get("name")
            if views_glob and not matches(view_name, views_glob):
                continue

            self._formatDescription(expanded)
            self.views.append(expanded)



#!/usr/bin/env python
# Copyright (C) 2015 Wayne Warren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import six


@six.add_metaclass(abc.ABCMeta)
class BaseSubCommand(object):
    """Base class for Jenkins Job Builder subcommands, intended to allow
    subcommands to be loaded as stevedore extensions by third party users.
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def parse_args(self, subparsers, recursive_parser):
        """Define subcommand arguments.

        :param subparsers
          A sub parser object. Implementations of this method should
          create a new subcommand parser by calling
            parser = subparsers.add_parser('command-name', ...)
          This will return a new ArgumentParser object; all other arguments to
          this method will be passed to the argparse.ArgumentParser constructor
          for the returned object.
        """

    @abc.abstractmethod
    def execute(self, config):
        """Execute subcommand behavior.

        :param config
          JJBConfig object containing final configuration from config files,
          command line arguments, and environment variables.
        """

    @staticmethod
    def parse_option_recursive_exclude(parser):
        """Add '--recursive'  and '--exclude' arguments to given parser."""
        parser.add_argument(
            "-r",
            "--recursive",
            action="store_true",
            dest="recursive",
            default=False,
            help="look for yaml files recursively",
        )

        parser.add_argument(
            "-x",
            "--exclude",
            dest="exclude",
            action="append",
            default=[],
            help="paths to exclude when using recursive search, "
            "uses standard globbing.",
        )



# Copyright 2015 Openstack Foundation

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The view list module handles creating Jenkins List views.

To create a list view specify ``list`` in the ``view-type`` attribute
to the :ref:`view_list` definition.

:View Parameters:
    * **name** (`str`): The name of the view.
    * **view-type** (`str`): The type of view.
    * **description** (`str`): A description of the view. (default '')
    * **filter-executors** (`bool`): Show only executors that can
      execute the included views. (default false)
    * **filter-queue** (`bool`): Show only included jobs in builder
      queue. (default false)
    * **job-name** (`list`): List of jobs to be included.
    * **job-filters** (`dict`): Job filters to be included. Requires
      :jenkins-plugins:`View Job Filters <view-job-filters>`

        * **most-recent** (`dict`)
            :most-recent:
                * **max-to-include** (`int`): Maximum number of jobs
                  to include. (default 0)
                * **check-start-time** (`bool`): Check job start
                  time. (default false)

        * **build-duration** (`dict`)
            :build-duration:
                * **match-type** ('str'): Jobs that match a filter
                  to include. (default includeMatched)
                * **build-duration-type** ('str'): Duration of the
                  build. (default Latest)
                * **amount-type**: ('str'): Duration in hours,
                  days or builds. (default Hours)
                * **amount**: ('int'): How far back to check.
                  (default 0)
                * **less-than**: ('bool'): Check build duration
                  less than or more than. (default True)
                * **build-duration-minutes**: ('int'): Build
                  duration minutes. (default 0)

        * **build-trend** (`dict`)
            :build-trend:
                * **match-type** ('str'): Jobs that match a filter
                  to include. (default includeMatched)
                * **build-trend-type** ('str'): Duration of the
                  build. (default Latest)
                * **amount-type**: ('str'): Duration in hours,
                  days or builds. (default Hours)
                * **amount**: ('int'): How far back to check.
                  (default 0)
                * **status**: ('str'): Job status.
                  (default Completed)

        * **job-status** (`dict`)
            :job-status:
                * **match-type** ('str'): Jobs that match a filter
                  to include. (default includeMatched)
                * **unstable** ('bool'): Jobs with status
                  unstable. (default False)
                * **failed** ('bool'): Jobs with status
                  failed. (default False)
                * **aborted** ('bool'): Jobs with status
                  aborted. (default False)
                * **disabled** ('bool'): Jobs with status
                  disabled. (default False)
                * **stable** ('bool'): Jobs with status
                  stable. (default False)

        * **fallback** (`dict`)
            :fallback:
                * **fallback-type** ('str'): Fallback type to include/exclude
                  for all jobs in a view, if no jobs have been included by
                  previous filters. (default REMOVE_ALL_IF_ALL_INCLUDED)

        * **build-status** (`dict`)
            :build-status:
                * **match-type** ('str'): Jobs that match a filter
                  to include. (default includeMatched)
                * **never-built** ('bool'): Jobs that are never
                  built. (default False)
                * **building** ('bool'): Jobs that are being
                  built. (default False)
                * **in-build-queue** ('bool'): Jobs that are in
                  the build queue. (default False)

        * **user-relevence** (`dict`)
            :user-relevence:
                * **match-type** ('str'): Jobs that match a filter
                  to include. (default includeMatched)
                * **build-count** ('str'): Count of builds.
                  (default AtLeastOne)
                * **amount-type**: ('str'): Duration in hours,
                  days or builds. (default Hours)
                * **amount**: ('int'): How far back to check.
                  (default 0)
                * **match-user-id** ('bool'): Jobs matching
                  user-id. (default False)
                * **match-user-fullname** ('bool'): Jobs
                  matching user fullname. (default False)
                * **ignore-case** ('bool'): Ignore case.
                  (default False)
                * **ignore-whitespace** ('bool'): Ignore
                  whitespace. (default False)
                * **ignore-non-alphaNumeric** ('bool'): Ignore
                  non-alphaNumeric. (default False)
                * **match-builder** ('bool'): Jobs matching
                  builder. (default False)
                * **match-email** ('bool'): Jobs matching
                  email. (default False)
                * **match-scm-changes** ('bool'): Jobs matching
                  scm changes. (default False)

        * **regex-job** (`dict`)
            :regex-job:
                * **match-type** ('str'): Jobs that match a filter
                  to include. (default includeMatched)
                * **regex-name** ('str'): Regular expression name.
                  (default '')
                * **regex** ('str'): Regular expression. (default '')

        * **job-tpye** (`dict`)
            :job-type:
                * **match-type** ('str'): Jobs that match a filter to include.
                  (default includeMatched)
                * **job-type** ('str'): Type of Job.
                  (default hudson.model.FreeStyleProject)

        * **parameter** (`dict`)
            :parameter:
                * **match-type** ('str'): Jobs that match a filter to include.
                  (default includeMatched)
                * **name** ('str'): Job name to match. (default '')
                * **value** ('str'): Value to match. (default '')
                * **desc** ('str'): Description to match. (default '')
                * **use-default-value** ('bool'): Use default value.
                  (default False)
                * **match-builds-in-progress** ('bool'): Match build in
                  progress. (default False)
                * **match-all-builds** ('bool'): Match all builds.
                  (default False)
                * **max-builds-to-match** ('int'): Maximum builds to match.
                  (default 0)

        * **other-views** (`dict`)
            :other-views:
                * **match-type** ('str'): Jobs that match a filter
                  to include. (default includeMatched)
                * **view-name** ('str'): View name.
                  (default select a view other than this one)

        * **scm** (`dict`)
            :scm:
                * **match-type** ('str'): Jobs that match a filter to include.
                  (default includeMatched)
                * **scm-type** ('str'): Type of SCM.
                  (default hudson.scm.NullSCM)

        * **secured-job** (`dict`)
            :secured-job:
                * **match-type** ('str'): Jobs that match a filter
                  to include. (default includeMatched)

        * **user-permissions** (`dict`)
            :user-permissions:
                * **match-type** ('str'): Jobs that match a filter to include.
                  (default includeMatched)
                * **configure** ('bool'): User with configure permissions.
                  (default false)
                * **amount-type**: ('bool'): User with build permissions.
                  (default false)
                * **amount**: ('bool'): User with workspace permissions.
                  (default false)
                * **permission-check**: ('str'): Match user permissions.
                  (default MustMatchAll)

        * **upstream-downstream** (`dict`)
            :upstream-downstream:
                * **include-upstream** ('bool'): Jobs that match upstream.
                  (default False)
                * **include-downstream** ('bool'): Jobs that match downstream.
                  (default False)
                * **recursive** ('bool'): Jobs that are recursive.
                  (default False)
                * **exclude-originals** ('bool'): Jobs that are originals.
                  (default False)

        * **unclassified** (`dict`)
            :unclassified:
                * **match-type** ('str'): Jobs that match a filter to include.
                  (default includeMatched)

    * **columns** (`list`): List of columns to be shown in view.
    * **regex** (`str`): . Regular expression for selecting jobs
      (optional)
    * **recurse** (`bool`): Recurse in subfolders.(default false)
    * **status-filter** (`bool`): Filter job list by enabled/disabled
      status. (optional)

Example:

    .. literalinclude::
        /../../tests/views/fixtures/view_list001.yaml

Example:

    .. literalinclude::
        /../../tests/views/fixtures/view_list002.yaml
"""

import xml.etree.ElementTree as XML
import jenkins_jobs.modules.base
import jenkins_jobs.modules.helpers as helpers
import jenkins_jobs.modules.view_jobfilters as view_jobfilters

COLUMN_DICT = {
    "status": "hudson.views.StatusColumn",
    "weather": "hudson.views.WeatherColumn",
    "job": "hudson.views.JobColumn",
    "last-success": "hudson.views.LastSuccessColumn",
    "last-failure": "hudson.views.LastFailureColumn",
    "last-duration": "hudson.views.LastDurationColumn",
    "build-button": "hudson.views.BuildButtonColumn",
    "last-stable": "hudson.views.LastStableColumn",
    "robot-list": "hudson.plugins.robot.view.RobotListViewColumn",
    "find-bugs": "hudson.plugins.findbugs.FindBugsColumn",
    "jacoco": "hudson.plugins.jacococoveragecolumn.JaCoCoColumn",
    "git-branch": "hudson.plugins.git.GitBranchSpecifierColumn",
    "favorite": "hudson.plugins.favorite.column.FavoriteColumn",
    "schedule-build": "org.jenkinsci.plugins.schedulebuild.ScheduleBuildButtonColumn",
    "priority-sorter": "jenkins.advancedqueue.PrioritySorterJobColumn",
    "build-filter": "hudson.views.BuildFilterColumn",
    "desc": "jenkins.branch.DescriptionColumn",
    "policy-violations": "com.sonatype.insight.ci.hudson.QualityColumn "
    'plugin="sonatype-clm-ci"',
    "member-graph-view": "com.barchart.jenkins.cascade.GraphViewColumn "
    'plugin="maven-release-cascade"',
    "built-on": [
        ["org.jenkins.plugins.builton.BuiltOnColumn", {"plugin": "built-on-column"}]
    ],
    "extra-tests-total": [
        ["jenkins.plugins.extracolumns.TestResultColumn", {"plugin": "extra-columns"}],
        "<testResultFormat>2</testResultFormat>",
    ],
    "extra-tests-failed": [
        ["jenkins.plugins.extracolumns.TestResultColumn", {"plugin": "extra-columns"}],
        "<testResultFormat>3</testResultFormat>",
    ],
    "extra-tests-passed": [
        ["jenkins.plugins.extracolumns.TestResultColumn", {"plugin": "extra-columns"}],
        "<testResultFormat>4</testResultFormat>",
    ],
    "extra-tests-skipped": [
        ["jenkins.plugins.extracolumns.TestResultColumn", {"plugin": "extra-columns"}],
        "<testResultFormat>5</testResultFormat>",
    ],
    "extra-tests-format-0": [
        ["jenkins.plugins.extracolumns.TestResultColumn", {"plugin": "extra-columns"}],
        "<testResultFormat>0</testResultFormat>",
    ],
    "extra-tests-format-1": [
        ["jenkins.plugins.extracolumns.TestResultColumn", {"plugin": "extra-columns"}],
        "<testResultFormat>1</testResultFormat>",
    ],
    "extra-build-description": [
        [
            "jenkins.plugins.extracolumns.BuildDescriptionColumn",
            {"plugin": "extra-columns"},
        ],
        "<columnWidth>3</columnWidth>",
        "<forceWidth>false</forceWidth>",
    ],
    "extra-build-parameters": [
        [
            "jenkins.plugins.extracolumns.BuildParametersColumn",
            {"plugin": "extra-columns"},
        ],
        "<singlePara>false</singlePara>",
        "<parameterName/>",
    ],
    "extra-last-user-name": "jenkins.plugins.extracolumns.UserNameColumn"
    ' plugin="extra-columns"',
    "extra-last-output": "jenkins.plugins.extracolumns.LastBuildConsoleColumn"
    ' plugin="extra-columns"',
    "extra-workspace-link": "jenkins.plugins.extracolumns.WorkspaceColumn "
    'plugin="extra-columns"',
    "extra-configure-button": "jenkins.plugins.extracolumns.ConfigureProjectColumn"
    ' plugin="extra-columns"',
}
DEFAULT_COLUMNS = [
    "status",
    "weather",
    "job",
    "last-success",
    "last-failure",
    "last-duration",
    "build-button",
]


class List(jenkins_jobs.modules.base.Base):
    sequence = 0

    def root_xml(self, data, sectioned=False):
        name = "hudson.model.ListView"
        if sectioned:
            name = "hudson.plugins.sectioned__view.ListViewSection"
        root = XML.Element(name)

        mapping = [("name", "name", None)]
        if not sectioned:
            mapping.extend(
                [
                    ("description", "description", ""),
                    ("filter-executors", "filterExecutors", False),
                    ("filter-queue", "filterQueue", False),
                ]
            )
        helpers.convert_mapping_to_xml(root, data, mapping, fail_required=True)

        if not sectioned:
            XML.SubElement(
                root, "properties", {"class": "hudson.model.View$PropertyList"}
            )

        jn_xml = XML.SubElement(root, "jobNames")
        jobnames = data.get("job-name", None)
        XML.SubElement(
            jn_xml, "comparator", {"class": "hudson.util.CaseInsensitiveComparator"}
        )
        if jobnames is not None:
            # Job names must be sorted in the xml
            jobnames = sorted(jobnames, key=str.lower)
            for jobname in jobnames:
                XML.SubElement(jn_xml, "string").text = str(jobname)

        job_filter_xml = XML.SubElement(root, "jobFilters")
        jobfilters = data.get("job-filters", [])

        for jobfilter in jobfilters:
            filter = getattr(view_jobfilters, jobfilter.replace("-", "_"))
            filter(job_filter_xml, jobfilters.get(jobfilter))

        if sectioned:
            mapping = [
                ("width", "width", "FULL", ["FULL", "HALF", "THIRD", "TWO_THIRDS"]),
                ("alignment", "alignment", "CENTER", ["CENTER", "LEFT", "RIGHT"]),
            ]
            helpers.convert_mapping_to_xml(root, data, mapping, fail_required=True)

        c_xml = XML.SubElement(root, "columns")
        columns = data.get("columns", DEFAULT_COLUMNS)

        for column in columns:
            if isinstance(column, dict):
                if "extra-build-parameter" in column:
                    p_name = column["extra-build-parameter"]
                    x = XML.SubElement(
                        c_xml,
                        "jenkins.plugins.extracolumns.BuildParametersColumn",
                        plugin="extra-columns",
                    )
                    x.append(XML.fromstring("<singlePara>true</singlePara>"))
                    x.append(
                        XML.fromstring("<parameterName>%s</parameterName>" % p_name)
                    )
            else:
                if column in COLUMN_DICT:
                    if isinstance(COLUMN_DICT[column], list):
                        x = XML.SubElement(
                            c_xml,
                            COLUMN_DICT[column][0][0],
                            **COLUMN_DICT[column][0][1]
                        )
                        for tag in COLUMN_DICT[column][1:]:
                            x.append(XML.fromstring(tag))
                    else:
                        XML.SubElement(c_xml, COLUMN_DICT[column])
        mapping = [("regex", "includeRegex", None)]
        if not sectioned:
            mapping.extend(
                [("recurse", "recurse", False), ("status-filter", "statusFilter", None)]
            )
        helpers.convert_mapping_to_xml(root, data, mapping, fail_required=False)

        return root



#!/usr/bin/env python
# Copyright (C) 2015 Wayne Warren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import os

import jenkins_jobs.version

from stevedore import extension


def __version__():
    return (
        "Jenkins Job Builder version: %s"
        % jenkins_jobs.version.version_info.version_string()
    )


def create_parser():
    """Create an ArgumentParser object usable by JenkinsJobs."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--conf",
        dest="conf",
        default=os.environ.get("JJB_CONF", None),
        help="configuration file [JJB_CONF]",
    )
    parser.add_argument(
        "-l",
        "--log_level",
        dest="log_level",
        default=os.environ.get("JJB_LOG_LEVEL", "info"),
        help="log level (default: %(default)s) [JJB_LOG_LEVEL]",
    )
    parser.add_argument(
        "--ignore-cache",
        action="store_true",
        dest="ignore_cache",
        default=None,
        help="ignore the cache and update the jobs anyhow (that will "
        "only flush the specified jobs cache)",
    )
    parser.add_argument(
        "--flush-cache",
        action="store_true",
        dest="flush_cache",
        default=None,
        help="flush all the cache entries before updating",
    )
    parser.add_argument(
        "--version",
        dest="version",
        action="version",
        version=__version__(),
        help="show version",
    )
    parser.add_argument(
        "--allow-empty-variables",
        action="store_true",
        dest="allow_empty_variables",
        default=None,
        help="Don't fail if any of the variables inside any string are "
        "not defined, replace with empty string instead.",
    )
    parser.add_argument(
        "--server",
        "-s",
        dest="section",
        default=os.environ.get("JJB_SECTION", "jenkins"),
        help="The Jenkins server ini section to use. Defaults to 'jenkins' "
        "[JJB_SECTION]",
    )
    parser.add_argument(
        "--user",
        "-u",
        default=os.environ.get("JJB_USER", None),
        help="The Jenkins user to use for authentication. This overrides "
        "the user specified in the configuration file. [JJB_USER]",
    )
    parser.add_argument(
        "--password",
        "-p",
        default=os.environ.get("JJB_PASSWORD", None),
        help="Password or API token to use for authenticating towards Jenkins."
        " This overrides the password specified in the configuration file."
        " [JJB_PASSWORD]",
    )

    subparser = parser.add_subparsers(
        dest="command", help="update, test, list or delete job"
    )

    extension_manager = extension.ExtensionManager(
        namespace="jjb.cli.subcommands", invoke_on_load=True
    )

    def parse_subcommand_args(ext, subparser):
        ext.obj.parse_args(subparser)

    extension_manager.map(parse_subcommand_args, subparser)

    return parser



# Copyright 2020 Openstack Foundation

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The view delivery pipeline module handles creation of Delivery Pipeline views.
To create a delivery pipeline view specify ``delivery_pipeline`` in the
``view-type`` attribute to the :ref:`view_delivery_pipeline` definition.
Requires the Jenkins :jenkins-plugins:`Delivery Pipeline Plugin
<delivery-pipeline-plugin>`.

:View Parameters:
    * **name** (`str`): The name of the view.
    * **view-type** (`str`): The type of view.
    * **description** (`str`): A description of the view. (optional)
    * **filter-executors** (`bool`): Show only executors that can
      execute the included views. (default false)
    * **filter-queue** (`bool`): Show only included jobs in builder
      queue. (default false)
    * **components** (`list`):
        * **name** (`str`): Name of the pipeline, usually the name of the
          component or product.
        * **initial-job** (`str`): First job in the pipeline.
        * **final-job** (`str`): Final job to display in the pipeline view
          regardless of its downstream jobs. (default '')
        * **show-upstream** (`bool`): Whether to show upstream. (default false)
    * **regexps** (`list`):
        * **regexp** (`str`): Regular expression to find initial jobs.
        * **show-upstream** (`bool`): Whether to show upstream. (default false)
    * **aggregated-changes-grouping-pattern** (`str`): Group changelog by regex
      pattern. (default '')
    * **allow-abort** (`bool`): Allow cancelling a running job from the
      delivery pipeline view. (default false)
    * **allow-manual-triggers** (`bool`): Displays a button in the pipeline
      view if a task is manual (Build other projects (manual step)) from Build
      Pipeline Plugin. (default false)
    * **allow-pipeline-start** (`bool`): Allow starting a new pipeline run from
      the delivery pipeline view. (default false)
    * **allow-rebuild** (`bool`): Allow rerunning a task from the delivery
      pipeline view. (default false)
    * **link-relative** (`bool`): Use relative links for jobs in this pipeline
      view to allow for easier navigation. (default false)
    * **link-to-console-log** (`bool`): Changes behaviour of task link in
      delivery pipeline view to go directly to the console log. (default false)
    * **max-number-of-visible-pipelines** (`int`): Limits the number of
      pipelines shown in the view, regardless of how many pipelines are
      configured. A negative value will not enforce a limit.
    * **no-of-columns** (`int`): Number of columns used for showing pipelines.
      Possible values are 1 (default), 2 and 3.
    * **no-of-pipelines** (`int`): Number of pipelines instances shown for each
      pipeline. Possible values are numbers from 1 to 50 (default 3).
    * **paging-enabled** (`bool`): Enable pagination in normal view, to allow
      navigation to older pipeline runs which are not displayed on the first
      page. Not available in full screen view. (default false)
    * **show-absolute-date-time** (`bool`): Show dates and times as absolute
      values instead of as relative to the current time. (default false)
    * **show-aggregated-changes** (`bool`): Show an aggregated changelog
      between different stages. (default false)
    * **show-aggregated-pipeline** (`bool`): Show an aggregated view where each
      stage shows the latest version being executed. (default false)
    * **show-avatars** (`bool`): Show avatars pictures instead of names of the
      people involved in a pipeline instance. (default false)
    * **show-changes** (`bool`): Show SCM change log for the first job in the
      pipeline. (default false)
    * **show-description** (`bool`): Show a build description connected to a
      specific pipeline task. (default false)
    * **show-promotions** (`bool`): Show promotions from Promoted Builds
      Plugin. (default false)
    * **show-static-analysis-results** (`bool`): Show different analysis
      results from Analysis Collector Plugin. (default false)
    * **show-test-results** (`bool`): Show test results as pass/failed/skipped.
      (default false)
    * **show-total-build-time** (`bool`): Show total build time for a pipeline
      run. (default false)
    * **sorting** (`str`): How to sort the pipelines in the current view. Only
      applicable when multiple pipelines are configured in the same view.
      Possible values are 'none' (default), 'title' (sort by title),
      'failed_last_activity' (sort by failed pipelines, then by last activity),
      'last_activity' (sort by last activity).
    * **update-interval** (`int`): How often the pipeline view will be updated.
      To be specified in seconds. (default 2)

Minimal Example:

    .. literalinclude::
        /../../tests/views/fixtures/view_delivery_pipeline-minimal.yaml

Full Example:

    .. literalinclude::
        /../../tests/views/fixtures/view_delivery_pipeline-full.yaml
"""

import xml.etree.ElementTree as XML
import jenkins_jobs.modules.base
import jenkins_jobs.modules.helpers as helpers


class DeliveryPipeline(jenkins_jobs.modules.base.Base):
    def root_xml(self, data):
        root = XML.Element(
            "se.diabol.jenkins.pipeline.DeliveryPipelineView",
            {"plugin": "delivery-pipeline-plugin"},
        )

        # Optional
        mapping_optional = [
            ("description", "description", None),
            ("filter-executors", "filterExecutors", False),
            ("filter-queue", "filterQueue", False),
        ]
        helpers.convert_mapping_to_xml(
            root, data, mapping_optional, fail_required=False
        )

        # Required - simple
        mapping = [
            (
                "aggregated-changes-grouping-pattern",
                "aggregatedChangesGroupingPattern",
                "",
            ),
            ("allow-abort", "allowAbort", False),
            ("allow-manual-triggers", "allowManualTriggers", False),
            ("allow-pipeline-start", "allowPipelineStart", False),
            ("allow-rebuild", "allowRebuild", False),
            ("link-relative", "linkRelative", False),
            ("link-to-console-log", "linkToConsoleLog", False),
            ("max-number-of-visible-pipelines", "maxNumberOfVisiblePipelines", -1),
            ("name", "name", None),
            ("no-of-columns", "noOfColumns", 1, [1, 2, 3]),
            ("no-of-pipelines", "noOfPipelines", 3, list(range(51))),
            ("paging-enabled", "pagingEnabled", False),
            ("show-absolute-date-time", "showAbsoluteDateTime", False),
            ("show-aggregated-changes", "showAggregatedChanges", False),
            ("show-aggregated-pipeline", "showAggregatedPipeline", False),
            ("show-avatars", "showAvatars", False),
            ("show-changes", "showChanges", False),
            ("show-description", "showDescription", False),
            ("show-promotions", "showPromotions", False),
            ("show-static-analysis-results", "showStaticAnalysisResults", False),
            ("show-test-results", "showTestResults", False),
            ("show-total-build-time", "showTotalBuildTime", False),
            ("update-interval", "updateInterval", 2),
        ]
        helpers.convert_mapping_to_xml(root, data, mapping, fail_required=True)

        # Required - complex
        sorting_val = data.get("sorting", "none")
        sorting_map = {
            "none": "None",
            "title": ("se.diabol.jenkins.pipeline.sort.NameComparator"),
            "failed_last_activity": (
                "se.diabol.jenkins.pipeline.sort.FailedJobComparator"
            ),
            "last_activity": (
                "se.diabol.jenkins.pipeline.sort.LatestActivityComparator"
            ),
        }
        sorting = XML.SubElement(root, "sorting")

        if sorting_val in sorting_map:
            sorting.text = sorting_map[sorting_val]
        else:
            sorting.text = sorting_map["none"]

        components = data.get("components", [])

        if len(components):
            component_specs = XML.SubElement(root, "componentSpecs")

            for c in components:
                component_spec = XML.SubElement(
                    component_specs,
                    "se.diabol.jenkins.pipeline.DeliveryPipelineView_-ComponentSpec",
                )

                name = XML.SubElement(component_spec, "name")
                name.text = c.get("name", "")

                first_job = XML.SubElement(component_spec, "firstJob")
                first_job.text = c.get("initial-job", "")

                last_job = XML.SubElement(component_spec, "lastJob")
                last_job.text = c.get("final-job", "")

                show_upstream = XML.SubElement(component_spec, "showUpstream")
                show_upstream.text = c.get("show-upstream", "false")

        regexps = data.get("regexps", [])

        if len(regexps):
            regexp_first_jobs = XML.SubElement(root, "regexpFirstJobs")

            for r in regexps:
                regexp_first_job = XML.SubElement(
                    regexp_first_jobs,
                    "se.diabol.jenkins.pipeline.DeliveryPipelineView_-RegExpSpec",
                )

                regexp = XML.SubElement(regexp_first_job, "regexp")
                regexp.text = r.get("regexp", "")

                show_upstream = XML.SubElement(regexp_first_job, "showUpstream")
                show_upstream.text = r.get("show-upstream", "false")

        return root



#!/usr/bin/env python
# Copyright (C) 2018 Sorin Sbarnea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import io
import os

from testscenarios.testcase import TestWithScenarios

from tests.base import mock
from tests.cmd.test_cmd import CmdTestsBase


@mock.patch("jenkins_jobs.builder.JenkinsManager.get_plugins_info", mock.MagicMock)
class ListFromJenkinsTests(TestWithScenarios, CmdTestsBase):

    scenarios = [
        ("single", dict(jobs=["job1"], globs=[], found=["job1"])),
        ("multiple", dict(jobs=["job1", "job2"], globs=[], found=["job1", "job2"])),
        (
            "multiple_with_glob",
            dict(
                jobs=["job1", "job2", "job3"],
                globs=["job[1-2]"],
                found=["job1", "job2"],
            ),
        ),
        (
            "multiple_with_multi_glob",
            dict(
                jobs=["job1", "job2", "job3", "job4"],
                globs=["job1", "job[24]"],
                found=["job1", "job2", "job4"],
            ),
        ),
    ]

    @mock.patch("jenkins_jobs.builder.JenkinsManager.get_jobs")
    def test_list(self, get_jobs_mock):
        def _get_jobs():
            return [{"name": name} for name in self.jobs]

        get_jobs_mock.side_effect = _get_jobs
        console_out = io.BytesIO()

        args = ["--conf", self.default_config_file, "list"] + self.globs

        with mock.patch("sys.stdout", console_out):
            self.execute_jenkins_jobs_with_args(args)

        self.assertEqual(
            console_out.getvalue().decode("utf-8").rstrip(), ("\n".join(self.found))
        )


@mock.patch("jenkins_jobs.builder.JenkinsManager.get_plugins_info", mock.MagicMock)
class ListFromYamlTests(TestWithScenarios, CmdTestsBase):

    scenarios = [
        ("all", dict(globs=[], found=["bam001", "bar001", "bar002", "baz001"])),
        (
            "some",
            dict(
                globs=["*am*", "*002", "bar001"], found=["bam001", "bar001", "bar002"]
            ),
        ),
    ]

    def test_list(self):
        path = os.path.join(self.fixtures_path, "cmd-002.yaml")

        console_out = io.BytesIO()
        with mock.patch("sys.stdout", console_out):
            self.execute_jenkins_jobs_with_args(
                ["--conf", self.default_config_file, "list", "-p", path] + self.globs
            )

        self.assertEqual(
            console_out.getvalue().decode("utf-8").rstrip(), ("\n".join(self.found))
        )



# Copyright 2012 Julian Taylor <jtaylor.debian@googlemail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
The matrix project module handles creating Jenkins matrix
projects. To create a matrix project specify ``matrix`` in the
``project-type`` attribute to the :ref:`Job` definition.
Currently it supports four axes which share the same
internal YAML structure:

* label expressions (``label-expression``)
* user-defined values (``user-defined``)
* slave name or label (``slave``)
* JDK name (``jdk``)

Requires the Jenkins :jenkins-plugins:`Matrix Project Plugin
<matrix-project>`.

The module also supports additional, plugin-defined axes:

* DynamicAxis (``dynamic``), requires the Jenkins
  :jenkins-plugins:`DynamicAxis Plugin <dynamic-axis>`
* GroovyAxis (``groovy``), requires the Jenkins
  :jenkins-wiki:`GroovyAxis Plugin <GroovyAxis>`
* YamlAxis (``yaml``), requires the Jenkins
  :jenkins-plugins:`Yaml Axis Plugin <yaml-axis>`

To tie the parent job to a specific node, you should use ``node`` parameter.
On a matrix project, this will tie *only* the parent job.  To restrict axes
jobs, you can define a single value ``slave`` axis.

:Job Parameters:

    .. note::

       You can only pick one of the strategies.

    * **execution-strategy** (optional, built in Jenkins):
        * **combination-filter** (`str`): axes selection filter
        * **sequential** (`bool`): run builds sequentially (default false)
        * **touchstone** (optional):
            * **expr** (`str`) -- selection filter for the touchstone build
            * **result** (`str`) -- required result of the job: \
            stable (default) or unstable

    * **yaml-strategy** (optional, requires
      :jenkins-plugins:`Yaml Axis Plugin <yaml-axis>`):

        * **exclude-key** (`str`) -- top key containing exclusion rules
        * Either one of:
        * **filename** (`str`) -- Yaml file containing exclusions
        * **text** (`str`) -- Inlined Yaml. Should be literal
          ``text: | exclude:...``

    * **axes** (`list`):
        * **axis**:
            * **type** (`str`) -- axis type, must be either type defined by
              :jenkins-plugins:`Matrix Project Plugin <matrix-project>`
              (``label-expression``, ``user-defined``, ``slave`` or ``jdk``) or
              a type defined by a plugin (see top of this document for a list
              of supported plugins).
            * **name** (`str`) -- name of the axis
            * **values** (`list`) -- values of the axis

The module supports also ShiningPanda axes:

Example:

.. literalinclude::  /../../tests/general/fixtures/matrix-axis003.yaml

Requires the Jenkins :jenkins-plugins:`ShiningPanda Plugin <shiningpanda>`.

Example:

  .. literalinclude::  /../../tests/yamlparser/fixtures/project-matrix001.yaml
    :language: yaml

Examples for yaml axis:

  .. literalinclude::  /../../tests/general/fixtures/matrix-axis-yaml.yaml
    :language: yaml

  .. literalinclude::
     /../../tests/general/fixtures/matrix-axis-yaml-strategy-file.yaml
    :language: yaml

  .. literalinclude::
     /../../tests/general/fixtures/matrix-axis-yaml-strategy-inlined.yaml
    :language: yaml
"""

import xml.etree.ElementTree as XML

import jenkins_jobs.modules.base
from jenkins_jobs.errors import InvalidAttributeError
from jenkins_jobs.modules import hudson_model


class Matrix(jenkins_jobs.modules.base.Base):
    sequence = 0
    # List the supported Axis names in our configuration
    # and map them to the Jenkins XML element name.
    supported_axis = {
        "label-expression": "hudson.matrix.LabelExpAxis",
        "user-defined": "hudson.matrix.TextAxis",
        "slave": "hudson.matrix.LabelAxis",
        "jdk": "hudson.matrix.JDKAxis",
        "dynamic": "ca.silvermaplesolutions.jenkins.plugins.daxis.DynamicAxis",
        "python": "jenkins.plugins.shiningpanda.matrix.PythonAxis",
        "tox": "jenkins.plugins.shiningpanda.matrix.ToxAxis",
        "groovy": "org.jenkinsci.plugins.GroovyAxis",
        "yaml": "org.jenkinsci.plugins.yamlaxis.YamlAxis",
    }

    supported_strategies = {
        # Jenkins built-in, default
        "execution-strategy": "hudson.matrix.DefaultMatrixExecutionStrategyImpl",
        "yaml-strategy": "org.jenkinsci.plugins.yamlaxis.YamlMatrixExecutionStrategy",
        "p4-strategy": "org.jenkinsci.plugins.p4.matrix.MatrixOptions",
    }

    def root_xml(self, data):
        root = XML.Element("matrix-project")

        # Default to 'execution-strategy'
        strategies = [s for s in data.keys() if s.endswith("-strategy")] or [
            "execution-strategy"
        ]

        # Job can not have multiple strategies
        if len(strategies) > 1:
            raise ValueError(
                "matrix-project does not support multiple strategies. "
                "Given %s: %s" % (len(strategies), ", ".join(strategies))
            )
        strategy_name = strategies[0]

        if strategy_name not in self.supported_strategies:
            raise ValueError(
                "Given strategy %s. Only %s strategies are supported"
                % (strategy_name, self.supported_strategies.keys())
            )

        ex_r = XML.SubElement(
            root,
            "executionStrategy",
            {"class": self.supported_strategies[strategy_name]},
        )

        strategy = data.get(strategy_name, {})

        if strategy_name == "execution-strategy":
            if "combination-filter" in strategy:
                XML.SubElement(root, "combinationFilter").text = str(
                    strategy.get("combination-filter", "")
                ).rstrip()
            XML.SubElement(ex_r, "runSequentially").text = str(
                strategy.get("sequential", False)
            ).lower()
            if "touchstone" in strategy:
                XML.SubElement(ex_r, "touchStoneCombinationFilter").text = str(
                    strategy["touchstone"].get("expr", "")
                )

                threshold = strategy["touchstone"].get("result", "stable").upper()
                supported_thresholds = ("STABLE", "UNSTABLE")
                if threshold not in supported_thresholds:
                    raise InvalidAttributeError(
                        "touchstone", threshold, supported_thresholds
                    )

                # Web ui uses Stable but hudson.model.Result has Success
                if threshold == "STABLE":
                    threshold = "SUCCESS"

                t_r = XML.SubElement(ex_r, "touchStoneResultCondition")
                for sub_elem in ("name", "ordinal", "color"):
                    XML.SubElement(t_r, sub_elem).text = hudson_model.THRESHOLDS[
                        threshold
                    ][sub_elem]

        elif strategy_name == "yaml-strategy":
            filename = str(strategy.get("filename", ""))
            text = str(strategy.get("text", ""))
            exclude_key = str(strategy.get("exclude-key", ""))

            if bool(filename) == bool(text):  # xor with str
                raise ValueError(
                    "yaml-strategy must be given " 'either "filename" or "text"'
                )

            yamlType = (filename and "file") or (text and "text")
            XML.SubElement(ex_r, "yamlType").text = yamlType

            XML.SubElement(ex_r, "yamlFile").text = filename
            XML.SubElement(ex_r, "yamlText").text = text

            XML.SubElement(ex_r, "excludeKey").text = exclude_key

        elif strategy_name == "p4-strategy":
            XML.SubElement(ex_r, "runSequentially").text = str(
                strategy.get("sequential", False)
            ).lower()

            XML.SubElement(ex_r, "buildParent").text = str(
                strategy.get("build-parent", False)
            ).lower()

        ax_root = XML.SubElement(root, "axes")
        for axis_ in data.get("axes", []):
            axis = axis_["axis"]
            axis_type = axis["type"]
            if axis_type not in self.supported_axis:
                raise ValueError(
                    "Only %s axes types are supported" % self.supported_axis.keys()
                )
            axis_name = self.supported_axis.get(axis_type)
            lbl_root = XML.SubElement(ax_root, axis_name)
            name, values = axis.get("name", ""), axis.get("values", [""])
            if axis_type == "jdk":
                XML.SubElement(lbl_root, "name").text = "jdk"
            elif axis_type == "python":
                XML.SubElement(lbl_root, "name").text = "PYTHON"
            elif axis_type == "tox":
                XML.SubElement(lbl_root, "name").text = "TOXENV"
            else:
                XML.SubElement(lbl_root, "name").text = str(name)
            if axis_type != "groovy":
                v_root = XML.SubElement(lbl_root, "values")
            if axis_type == "dynamic":
                XML.SubElement(v_root, "string").text = str(values[0])
                XML.SubElement(lbl_root, "varName").text = str(values[0])
                v_root = XML.SubElement(lbl_root, "axisValues")
                XML.SubElement(v_root, "string").text = "default"
            elif axis_type == "groovy":
                command = XML.SubElement(lbl_root, "groovyString")
                command.text = axis.get("command")
                XML.SubElement(lbl_root, "computedValues").text = ""
            elif axis_type == "yaml":
                XML.SubElement(v_root, "string").text = axis.get("filename")
            else:
                for v in values:
                    XML.SubElement(v_root, "string").text = str(v)

        return root



#
# Copyright (c) 2016 Kien Ha <kienha9922@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from testtools.matchers import Equals
import xml.etree.ElementTree as XML
import yaml

from jenkins_jobs.errors import InvalidAttributeError
from jenkins_jobs.errors import MissingAttributeError
from jenkins_jobs.errors import JenkinsJobsException
from jenkins_jobs.modules.helpers import (
    convert_mapping_to_xml,
    check_mutual_exclusive_data_args,
)
from tests import base


class TestCaseTestHelpers(base.BaseTestCase):
    def test_convert_mapping_to_xml(self):
        """
        Tests the test_convert_mapping_to_xml_fail_required function
        """

        # Test default values
        default_root = XML.Element("testdefault")
        default_data = yaml.safe_load("string: hello")
        default_mappings = [("default-string", "defaultString", "default")]

        convert_mapping_to_xml(
            default_root, default_data, default_mappings, fail_required=True
        )
        result = default_root.find("defaultString").text
        self.assertThat(result, Equals("default"))

        # Test user input
        user_input_root = XML.Element("testUserInput")
        user_input_data = yaml.safe_load("user-input-string: hello")
        user_input_mappings = [("user-input-string", "userInputString", "user-input")]

        convert_mapping_to_xml(
            user_input_root, user_input_data, user_input_mappings, fail_required=True
        )
        result = user_input_root.find("userInputString").text
        self.assertThat(result, Equals("hello"))

        # Test missing required input
        required_root = XML.Element("testrequired")
        required_data = yaml.safe_load("string: hello")
        required_mappings = [("required-string", "requiredString", None)]

        self.assertRaises(
            MissingAttributeError,
            convert_mapping_to_xml,
            required_root,
            required_data,
            required_mappings,
            fail_required=True,
        )

        # Test invalid user input for list
        user_input_root = XML.Element("testUserInput")
        user_input_data = yaml.safe_load("user-input-string: bye")
        valid_inputs = ["hello"]
        user_input_mappings = [
            ("user-input-string", "userInputString", "user-input", valid_inputs)
        ]

        self.assertRaises(
            InvalidAttributeError,
            convert_mapping_to_xml,
            user_input_root,
            user_input_data,
            user_input_mappings,
        )

        # Test invalid user input for dict
        user_input_root = XML.Element("testUserInput")
        user_input_data = yaml.safe_load("user-input-string: later")
        valid_inputs = {"hello": "world"}
        user_input_mappings = [
            ("user-input-string", "userInputString", "user-input", valid_inputs)
        ]

        self.assertRaises(
            InvalidAttributeError,
            convert_mapping_to_xml,
            user_input_root,
            user_input_data,
            user_input_mappings,
        )

        # Test invalid key for dict
        user_input_root = XML.Element("testUserInput")
        user_input_data = yaml.safe_load("user-input-string: world")
        valid_inputs = {"hello": "world"}
        user_input_mappings = [
            ("user-input-string", "userInputString", "user-input", valid_inputs)
        ]

        self.assertRaises(
            InvalidAttributeError,
            convert_mapping_to_xml,
            user_input_root,
            user_input_data,
            user_input_mappings,
        )

    def test_check_mutual_exclusive_data_args_no_mutual_exclusive(self):
        @check_mutual_exclusive_data_args(0, "foo", "bar")
        @check_mutual_exclusive_data_args(0, "foo", "baz")
        def func(data):
            pass

        func({"baz": "qaz", "bar": "qaz"})

    def test_check_mutual_exclusive_data_args_mutual_exclusive(self):
        @check_mutual_exclusive_data_args(0, "foo", "bar")
        @check_mutual_exclusive_data_args(0, "foo", "baz")
        def func(data):
            pass

        self.assertRaises(JenkinsJobsException, func, {"foo": "qaz", "bar": "qaz"})



#!/usr/bin/env python
# Copyright (C) 2012 OpenStack, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Manage jobs in Jenkins server

import errno
import io
import logging
import os
import re
import tempfile

import fasteners
import yaml

from jenkins_jobs import errors

logger = logging.getLogger(__name__)


class JobCache(object):
    # ensure each instance of the class has a reference to the required
    # modules so that they are available to be used when the destructor
    # is being called since python will not guarantee that it won't have
    # removed global module references during teardown.
    _logger = logger
    _os = os
    _tempfile = tempfile
    _yaml = yaml

    def __init__(self, jenkins_url, flush=False):
        cache_dir = self.get_cache_dir()
        # One cache per remote Jenkins URL:
        host_vary = re.sub(r"[^A-Za-z0-9\-\~]", "_", jenkins_url)
        self.cachefilename = os.path.join(
            cache_dir, "cache-host-jobs-" + host_vary + ".yml"
        )

        # generate named lockfile if none exists, and lock it
        self._locked = self._lock()
        if not self._locked:
            raise errors.JenkinsJobsException(
                "Unable to lock cache for '%s'" % jenkins_url
            )

        if flush or not os.path.isfile(self.cachefilename):
            self.data = {}
        else:
            with io.open(self.cachefilename, "r", encoding="utf-8") as yfile:
                self.data = yaml.safe_load(yfile)
        logger.debug("Using cache: '{0}'".format(self.cachefilename))

    def _lock(self):
        self._fastener = fasteners.InterProcessLock("%s.lock" % self.cachefilename)

        return self._fastener.acquire(delay=1, max_delay=2, timeout=60)

    def _unlock(self):
        if getattr(self, "_locked", False):
            if getattr(self, "_fastener", None) is not None:
                self._fastener.release()
            self._locked = None

    @staticmethod
    def get_cache_dir():
        home = os.path.expanduser("~")
        if home == "~":
            raise OSError("Could not locate home folder")
        xdg_cache_home = os.environ.get("XDG_CACHE_HOME") or os.path.join(
            home, ".cache"
        )
        path = os.path.join(xdg_cache_home, "jenkins_jobs")
        if not os.path.isdir(path):
            try:
                os.makedirs(path)
            except OSError as ose:
                # it could happen that two jjb instances are running at the
                # same time and that the other instance created the directory
                # after we made the check, in which case there is no error
                if ose.errno != errno.EEXIST:
                    raise
        return path

    def set(self, job, md5):
        self.data[job] = md5

    def clear(self):
        self.data.clear()

    def is_cached(self, job):
        if job in self.data:
            return True
        return False

    def has_changed(self, job, md5):
        if job in self.data and self.data[job] == md5:
            return False
        return True

    def save(self):
        # use self references to required modules in case called via __del__
        # write to tempfile under same directory and then replace to avoid
        # issues around corruption such the process be killed
        tfile = self._tempfile.NamedTemporaryFile(
            dir=self.get_cache_dir(), delete=False
        )
        tfile.write(self._yaml.dump(self.data).encode("utf-8"))
        # force contents to be synced on disk before overwriting cachefile
        tfile.flush()
        self._os.fsync(tfile.fileno())
        tfile.close()
        try:
            self._os.rename(tfile.name, self.cachefilename)
        except OSError:
            # On Windows, if dst already exists, OSError will be raised even if
            # it is a file. Remove the file first in that case and try again.
            self._os.remove(self.cachefilename)
            self._os.rename(tfile.name, self.cachefilename)

        self._logger.debug("Cache written out to '%s'" % self.cachefilename)

    def __del__(self):
        # check we initialized sufficiently in case called
        # due to an exception occurring in the __init__
        if getattr(self, "data", None) is not None:
            try:
                self.save()
            except Exception as e:
                self._logger.error(
                    "Failed to write to cache file '%s' on "
                    "exit: %s" % (self.cachefilename, e)
                )
        self._unlock()



# Copyright 2012 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
The Maven Project module handles creating Maven Jenkins projects.

To create a Maven project, specify ``maven`` in the ``project-type``
attribute to the :ref:`Job` definition. It also requires a ``maven`` section
in the :ref:`Job` definition.

:Job Parameters:
    * **root-module**:
        * **group-id** (`str`): GroupId.
        * **artifact-id** (`str`): ArtifactId.
    * **root-pom** (`str`): The path to the pom.xml file. (default 'pom.xml')
    * **goals** (`str`): Goals to execute. (required)
    * **maven-opts** (`str`): Java options to pass to maven (aka MAVEN_OPTS)
    * **maven-name** (`str`): Installation of maven which should be used.
      Not setting ``maven-name`` appears to use the first maven install
      defined in the global jenkins config.
    * **private-repository** ('str'): Whether to use a private maven repository
      Possible values are `default`, `local-to-workspace` and
      `local-to-executor`.
    * **ignore-upstream-changes** (`bool`): Do not start a build whenever
      a SNAPSHOT dependency is built or not. (default true)
    * **incremental-build** (`bool`): Activate incremental build - only build
      changed modules (default false).
    * **automatic-archiving** (`bool`): Activate automatic artifact archiving
      (default true).
    * **automatic-site-archiving** (`bool`): Activate automatic site
      documentation artifact archiving (default true).
    * **automatic-fingerprinting** (`bool`): Activate automatic fingerprinting
      of consumed and produced artifacts (default true).
    * **per-module-email** (`bool`): Send an e-mail for each failed module
      (default true).
    * **parallel-build-modules** (`bool`): Build modules in parallel
      (default false)
    * **resolve-dependencies** (`bool`): Resolve Dependencies during Pom
      parsing (default false).
    * **run-headless** (`bool`): Run headless (default false).
    * **disable-downstream** (`bool`): Disable triggering of downstream
      projects (default false).
    * **process-plugins** (`bool`): Process Plugins during Pom parsing
      (default false).
    * **custom-workspace** (`str`): Path to the custom workspace. If no path is
      provided, custom workspace is not used. (optional)
    * **settings** (`str`): Path to custom maven settings file. If settings
      type is 'file' then this is a Path. Otherwise it is the id for
      ConfigFileProvider. (optional)
    * **settings-type** (`str`): Type of settings file file|cfp.
      (default file)
    * **global-settings** (`str`): Path to custom maven global settings file.
      If settings type is 'file' then this is a Path. Otherwise it is the id
      for ConfigFileProvider. (optional)
    * **global-settings-type** (`str`): Type of settings file file|cfp.
      (default file)
    * **post-step-run-condition** (`str`): Run the post-build steps only if the
      build succeeds ('SUCCESS'), build succeeds or is unstable ('UNSTABLE'),
      regardless of build result ('FAILURE'). (default 'FAILURE').

Requires the Jenkins :jenkins-plugins:`Config File Provider Plugin
<config-file-provider>`
for the Config File Provider "settings" and "global-settings" config.

Example:

    .. literalinclude:: /../../tests/general/fixtures/project-maven001.yaml

CFP Example:

    .. literalinclude:: /../../tests/general/fixtures/project-maven003.yaml
"""
import pkg_resources
import xml.etree.ElementTree as XML

from jenkins_jobs.errors import InvalidAttributeError
import jenkins_jobs.modules.base
import jenkins_jobs.modules.helpers as helpers
from jenkins_jobs.modules import hudson_model


class Maven(jenkins_jobs.modules.base.Base):
    sequence = 0

    choices_private_repo = {
        "default": "hudson.maven.local_repo.DefaultLocalRepositoryLocator",
        "local-to-workspace": "hudson.maven.local_repo.PerJobLocalRepositoryLocator",
        "local-to-executor": "hudson.maven.local_repo.PerExecutorLocalRepositoryLocator",
    }

    def root_xml(self, data):
        xml_parent = XML.Element("maven2-moduleset")
        if "maven" not in data:
            return xml_parent

        # determine version of plugin
        plugin_info = self.registry.get_plugin_info("Maven Integration plugin")
        version = pkg_resources.parse_version(plugin_info.get("version", "0"))

        if "root-module" in data["maven"]:
            root_module = XML.SubElement(xml_parent, "rootModule")
            XML.SubElement(root_module, "groupId").text = data["maven"]["root-module"][
                "group-id"
            ]
            XML.SubElement(root_module, "artifactId").text = data["maven"][
                "root-module"
            ]["artifact-id"]
        XML.SubElement(xml_parent, "goals").text = data["maven"]["goals"]

        maven_opts = data["maven"].get("maven-opts")
        if maven_opts:
            XML.SubElement(xml_parent, "mavenOpts").text = maven_opts

        maven_name = data["maven"].get("maven-name")
        if maven_name:
            XML.SubElement(xml_parent, "mavenName").text = maven_name

        private_repo = data["maven"].get("private-repository")
        if private_repo:
            if private_repo not in self.choices_private_repo.keys():
                raise ValueError(
                    'Not a valid private-repository "%s", '
                    'must be one of "%s"'
                    % (private_repo, ", ".join(self.choices_private_repo.keys()))
                )
            XML.SubElement(
                xml_parent,
                "localRepository",
                attrib={"class": self.choices_private_repo[private_repo]},
            )

        XML.SubElement(xml_parent, "ignoreUpstremChanges").text = str(
            data["maven"].get("ignore-upstream-changes", True)
        ).lower()

        XML.SubElement(xml_parent, "rootPOM").text = data["maven"].get(
            "root-pom", "pom.xml"
        )
        XML.SubElement(xml_parent, "aggregatorStyleBuild").text = str(
            not data["maven"].get("parallel-build-modules", False)
        ).lower()
        XML.SubElement(xml_parent, "incrementalBuild").text = str(
            data["maven"].get("incremental-build", False)
        ).lower()
        XML.SubElement(xml_parent, "siteArchivingDisabled").text = str(
            not data["maven"].get("automatic-site-archiving", True)
        ).lower()
        XML.SubElement(xml_parent, "fingerprintingDisabled").text = str(
            not data["maven"].get("automatic-fingerprinting", True)
        ).lower()
        if version > pkg_resources.parse_version(
            "0"
        ) and version < pkg_resources.parse_version("2.0.1"):
            XML.SubElement(xml_parent, "perModuleEmail").text = str(
                data.get("per-module-email", True)
            ).lower()
        XML.SubElement(xml_parent, "archivingDisabled").text = str(
            not data["maven"].get("automatic-archiving", True)
        ).lower()
        XML.SubElement(xml_parent, "resolveDependencies").text = str(
            data["maven"].get("resolve-dependencies", False)
        ).lower()
        XML.SubElement(xml_parent, "processPlugins").text = str(
            data["maven"].get("process-plugins", False)
        ).lower()
        XML.SubElement(xml_parent, "mavenValidationLevel").text = "-1"
        XML.SubElement(xml_parent, "runHeadless").text = str(
            data["maven"].get("run-headless", False)
        ).lower()
        XML.SubElement(xml_parent, "disableTriggerDownstreamProjects").text = str(
            data["maven"].get("disable-downstream", False)
        ).lower()
        if "custom-workspace" in data["maven"]:
            XML.SubElement(xml_parent, "customWorkspace").text = str(
                data["maven"].get("custom-workspace")
            )
        helpers.config_file_provider_settings(xml_parent, data["maven"])

        run_post_steps = XML.SubElement(xml_parent, "runPostStepsIfResult")
        run_conditions = ["SUCCESS", "UNSTABLE", "FAILURE"]
        run_condition = data["maven"].get("post-step-run-condition", "FAILURE")
        if run_condition not in run_conditions:
            raise InvalidAttributeError(
                "post-step-run-condition", run_condition, run_conditions
            )
        cond_dict = hudson_model.THRESHOLDS[run_condition]
        XML.SubElement(run_post_steps, "name").text = cond_dict["name"]
        XML.SubElement(run_post_steps, "ordinal").text = cond_dict["ordinal"]
        XML.SubElement(run_post_steps, "color").text = cond_dict["color"]

        return xml_parent



# -*- coding: utf-8 -*-
# Copyright (C) 2015 David Caro <david@dcaro.es>
#
# Based on jenkins_jobs/modules/project_flow.py by
# Copyright (C) 2013 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Deprecated: please use :ref:`project_pipeline` instead.

The workflow Project module handles creating Jenkins workflow projects.
You may specify ``workflow`` in the ``project-type`` attribute of
the :ref:`Job` definition.
For now only inline scripts are supported.

Requires the Jenkins :jenkins-plugins:`Workflow Plugin <workflow-aggregator>`.

In order to use it for job-template you have to escape the curly braces by
doubling them in the DSL: { -> {{ , otherwise it will be interpreted by the
python str.format() command.

:Job Parameters:
    * **dsl** (`str`): The DSL content.
    * **sandbox** (`bool`): If the script should run in a sandbox (default
      false)

Job example:

    .. literalinclude::
      /../../tests/yamlparser/fixtures/project_workflow_template001.yaml

Job template example:

    .. literalinclude::
      /../../tests/yamlparser/fixtures/project_workflow_template002.yaml

"""
import logging
import xml.etree.ElementTree as XML

import jenkins_jobs.modules.base
import jenkins_jobs.modules.helpers as helpers


class Workflow(jenkins_jobs.modules.base.Base):
    sequence = 0

    def root_xml(self, data):
        logger = logging.getLogger(__name__)
        logger.warning("Workflow job type is deprecated, please use Pipeline job type")

        xml_parent = XML.Element("flow-definition", {"plugin": "workflow-job"})
        xml_definition = XML.SubElement(
            xml_parent,
            "definition",
            {
                "class": "org.jenkinsci.plugins." "workflow.cps.CpsFlowDefinition",
                "plugin": "workflow-cps",
            },
        )

        mapping = [("dsl", "script", None), ("sandbox", "sandbox", False)]
        helpers.convert_mapping_to_xml(
            xml_definition, data, mapping, fail_required=True
        )

        return xml_parent



#!/usr/bin/env python
# Copyright (C) 2018 Sorin Sbarnea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import logging
import sys
import jenkins_jobs.cli.subcommand.base as base
import jenkins_jobs.utils as utils
import jenkins_jobs.builder as builder
import jenkins_jobs.parser as parser
import jenkins_jobs.registry as registry


def list_duplicates(seq):
    seen = set()
    return set(x for x in seq if x in seen or seen.add(x))


class ListSubCommand(base.BaseSubCommand):
    def parse_args(self, subparser):
        list = subparser.add_parser("list", help="List jobs")

        self.parse_option_recursive_exclude(list)

        list.add_argument("names", help="name(s) of job(s)", nargs="*", default=None)
        list.add_argument(
            "-p", "--path", default=None, help="path to YAML file or directory"
        )

    def execute(self, options, jjb_config):
        self.jjb_config = jjb_config
        self.jenkins = builder.JenkinsManager(jjb_config)

        jobs = self.get_jobs(options.names, options.path)

        logging.info("Matching jobs: %d", len(jobs))
        stdout = utils.wrap_stream(sys.stdout)

        for job in jobs:
            stdout.write((job + "\n").encode("utf-8"))

    def get_jobs(self, jobs_glob=None, fn=None):
        if fn:
            r = registry.ModuleRegistry(self.jjb_config, self.jenkins.plugins_list)
            p = parser.YamlParser(self.jjb_config)
            p.load_files(fn)
            p.expandYaml(r, jobs_glob)
            jobs = [j["name"] for j in p.jobs]
        else:
            jobs = [
                j["name"]
                for j in self.jenkins.get_jobs()
                if not jobs_glob or parser.matches(j["name"], jobs_glob)
            ]

        jobs = sorted(jobs)
        for duplicate in list_duplicates(jobs):
            logging.warning("Found duplicate job name '%s', likely bug.", duplicate)

        logging.debug("Builder.get_jobs: returning %r", jobs)

        return jobs



# Copyright 2015 David Caro
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time
from multiprocessing import cpu_count

from testtools import matchers
from testtools import TestCase

from jenkins_jobs.parallel import concurrent
from tests.base import mock


class TestCaseParallel(TestCase):
    def test_parallel_correct_order(self):
        expected = list(range(10, 20))

        @concurrent
        def parallel_test(num_base, num_extra):
            return num_base + num_extra

        parallel_args = [{"num_extra": num} for num in range(10)]
        result = parallel_test(10, concurrent=parallel_args)
        self.assertThat(result, matchers.Equals(expected))

    def test_parallel_time_less_than_serial(self):
        @concurrent
        def wait(secs):
            time.sleep(secs)

        before = time.time()
        # ten threads to make it as fast as possible
        wait(concurrent=[{"secs": 1} for _ in range(10)], n_workers=10)
        after = time.time()
        self.assertThat(after - before, matchers.LessThan(5))

    def test_parallel_single_thread(self):
        expected = list(range(10, 20))

        @concurrent
        def parallel_test(num_base, num_extra):
            return num_base + num_extra

        parallel_args = [{"num_extra": num} for num in range(10)]
        result = parallel_test(10, concurrent=parallel_args, n_workers=1)
        self.assertThat(result, matchers.Equals(expected))

    @mock.patch("jenkins_jobs.parallel.cpu_count", wraps=cpu_count)
    def test_use_auto_detect_cores(self, mockCpu_count):
        @concurrent
        def parallel_test():
            return True

        result = parallel_test(concurrent=[{} for _ in range(10)], n_workers=0)
        self.assertThat(result, matchers.Equals([True for _ in range(10)]))
        mockCpu_count.assert_called_once_with()



#!/usr/bin/env python
#
# Copyright 2013 Darragh Bailey
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import yaml

from testtools import ExpectedException
from yaml.composer import ComposerError

from jenkins_jobs.config import JJBConfig
from jenkins_jobs.parser import YamlParser
from tests import base


def _exclude_scenarios(input_filename):
    return os.path.basename(input_filename).startswith("custom_")


class TestCaseLocalYamlInclude(base.JsonTestCase):
    """
    Verify application specific tags independently of any changes to
    modules XML parsing behaviour
    """

    fixtures_path = os.path.join(os.path.dirname(__file__), "fixtures")
    scenarios = base.get_scenarios(
        fixtures_path, "yaml", "json", filter_func=_exclude_scenarios
    )

    def test_yaml_snippet(self):

        if os.path.basename(self.in_filename).startswith("exception_"):
            with ExpectedException(ComposerError, "^found duplicate anchor .*"):
                super(TestCaseLocalYamlInclude, self).test_yaml_snippet()
        else:
            super(TestCaseLocalYamlInclude, self).test_yaml_snippet()


class TestCaseLocalYamlAnchorAlias(base.YamlTestCase):
    """
    Verify yaml input is expanded to the expected yaml output when using yaml
    anchors and aliases.
    """

    fixtures_path = os.path.join(os.path.dirname(__file__), "fixtures")
    scenarios = base.get_scenarios(fixtures_path, "iyaml", "oyaml")


class TestCaseLocalYamlIncludeAnchors(base.BaseTestCase):

    fixtures_path = os.path.join(os.path.dirname(__file__), "fixtures")

    def test_multiple_same_anchor_in_multiple_toplevel_yaml(self):
        """
        Verify that anchors/aliases only span use of '!include' tag

        To ensure that any yaml loaded by the include tag is in the same
        space as the top level file, but individual top level yaml definitions
        are treated by the yaml loader as independent.
        """

        files = [
            "custom_same_anchor-001-part1.yaml",
            "custom_same_anchor-001-part2.yaml",
        ]

        jjb_config = JJBConfig()
        jjb_config.jenkins["url"] = "http://example.com"
        jjb_config.jenkins["user"] = "jenkins"
        jjb_config.jenkins["password"] = "password"
        jjb_config.builder["plugins_info"] = []
        jjb_config.validate()
        j = YamlParser(jjb_config)
        j.load_files([os.path.join(self.fixtures_path, f) for f in files])


class TestCaseLocalYamlRetainAnchors(base.BaseTestCase):

    fixtures_path = os.path.join(os.path.dirname(__file__), "fixtures")

    def test_retain_anchors_default(self):
        """
        Verify that anchors are NOT retained across files by default.
        """

        files = ["custom_retain_anchors_include001.yaml", "custom_retain_anchors.yaml"]

        jjb_config = JJBConfig()
        # use the default value for retain_anchors
        jjb_config.validate()
        j = YamlParser(jjb_config)
        with ExpectedException(yaml.composer.ComposerError, "found undefined alias.*"):
            j.load_files([os.path.join(self.fixtures_path, f) for f in files])

    def test_retain_anchors_enabled(self):
        """
        Verify that anchors are retained across files if retain_anchors is
        enabled in the config.
        """

        files = ["custom_retain_anchors_include001.yaml", "custom_retain_anchors.yaml"]

        jjb_config = JJBConfig()
        jjb_config.yamlparser["retain_anchors"] = True
        jjb_config.validate()
        j = YamlParser(jjb_config)
        j.load_files([os.path.join(self.fixtures_path, f) for f in files])



#!/usr/bin/env python
# Copyright (C) 2015 OpenStack, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Manage interpolation of JJB variables into template strings.

import logging
from pprint import pformat
import re
from string import Formatter

from jenkins_jobs.errors import JenkinsJobsException
from jenkins_jobs.local_yaml import CustomLoader

logger = logging.getLogger(__name__)


def deep_format(obj, paramdict, allow_empty=False):
    """Deep format configuration.

    Apply the paramdict via str.format() to all string objects found within
    the supplied obj. Lists and dicts are traversed recursively.
    """
    # YAML serialisation was originally used to achieve this, but that places
    # limitations on the values in paramdict - the post-format result must
    # still be valid YAML (so substituting-in a string containing quotes, for
    # example, is problematic).
    if hasattr(obj, "format"):
        try:
            ret = CustomFormatter(allow_empty).format(obj, **paramdict)
        except KeyError as exc:
            missing_key = exc.args[0]
            desc = "%s parameter missing to format %s\nGiven:\n%s" % (
                missing_key,
                obj,
                pformat(paramdict),
            )
            raise JenkinsJobsException(desc)
        except Exception:
            logging.error(
                "Problem formatting with args:\nallow_empty:"
                "%s\nobj: %s\nparamdict: %s" % (allow_empty, obj, paramdict)
            )
            raise

    elif isinstance(obj, list):
        ret = type(obj)()
        for item in obj:
            ret.append(deep_format(item, paramdict, allow_empty))
    elif isinstance(obj, dict):
        ret = type(obj)()
        for item in obj:
            try:
                ret[deep_format(item, paramdict, allow_empty)] = deep_format(
                    obj[item], paramdict, allow_empty
                )
            except KeyError as exc:
                missing_key = exc.args[0]
                desc = "%s parameter missing to format %s\nGiven:\n%s" % (
                    missing_key,
                    obj,
                    pformat(paramdict),
                )
                raise JenkinsJobsException(desc)
            except Exception:
                logging.error(
                    "Problem formatting with args:\nallow_empty:"
                    "%s\nobj: %s\nparamdict: %s" % (allow_empty, obj, paramdict)
                )
                raise
    else:
        ret = obj
    if isinstance(ret, CustomLoader):
        # If we have a CustomLoader here, we've lazily-loaded a template
        # or rendered a template to a piece of YAML;
        # attempt to format it.
        ret = deep_format(
            ret.get_object_to_format(), paramdict, allow_empty=allow_empty
        )
    return ret


class CustomFormatter(Formatter):
    """
    Custom formatter to allow non-existing key references when formatting a
    string
    """

    _expr = r"""
        (?<!{){({{)*                # non-pair opening {
        (?:obj:)?                   # obj:
        (?P<key>\w+)                # key
        (?:\|(?P<default>[^}]*))?   # default fallback
        }(}})*(?!})                 # non-pair closing }
    """

    def __init__(self, allow_empty=False):
        super(CustomFormatter, self).__init__()
        self.allow_empty = allow_empty

    def vformat(self, format_string, args, kwargs):
        matcher = re.compile(self._expr, re.VERBOSE)

        # special case of returning the object if the entire string
        # matches a single parameter
        try:
            result = re.match("^%s$" % self._expr, format_string, re.VERBOSE)
        except TypeError:
            return format_string.format(**kwargs)
        if result is not None:
            try:
                return kwargs[result.group("key")]
            except KeyError:
                pass

        # handle multiple fields within string via a callback to re.sub()
        def re_replace(match):
            key = match.group("key")
            default = match.group("default")

            if default is not None:
                if key not in kwargs:
                    return default
                else:
                    return "{%s}" % key
            return match.group(0)

        format_string = matcher.sub(re_replace, format_string)

        return Formatter.vformat(self, format_string, args, kwargs)

    def get_value(self, key, args, kwargs):
        try:
            return Formatter.get_value(self, key, args, kwargs)
        except KeyError:
            if self.allow_empty:
                logger.debug(
                    "Found uninitialized key %s, replaced with empty string", key
                )
                return ""
            raise



import pkg_resources

from testtools.content import text_content
import testscenarios

from jenkins_jobs.config import JJBConfig
from jenkins_jobs.registry import ModuleRegistry
from tests import base


class ModuleRegistryPluginInfoTestsWithScenarios(
    testscenarios.TestWithScenarios, base.BaseTestCase
):
    scenarios = [
        ("s1", dict(v1="1.0.0", op="__gt__", v2="0.8.0")),
        ("s2", dict(v1="1.0.1alpha", op="__gt__", v2="1.0.0")),
        ("s3", dict(v1="1.0", op="__eq__", v2="1.0.0")),
        ("s4", dict(v1="1.0", op="__eq__", v2="1.0")),
        ("s5", dict(v1="1.0", op="__lt__", v2="1.8.0")),
        ("s6", dict(v1="1.0.1alpha", op="__lt__", v2="1.0.1")),
        ("s7", dict(v1="1.0alpha", op="__lt__", v2="1.0.0")),
        ("s8", dict(v1="1.0-alpha", op="__lt__", v2="1.0.0")),
        ("s9", dict(v1="1.1-alpha", op="__gt__", v2="1.0")),
        ("s10", dict(v1="1.0-SNAPSHOT", op="__lt__", v2="1.0")),
        ("s11", dict(v1="1.0.preview", op="__lt__", v2="1.0")),
        ("s12", dict(v1="1.1-SNAPSHOT", op="__gt__", v2="1.0")),
        ("s13", dict(v1="1.0a-SNAPSHOT", op="__lt__", v2="1.0a")),
        (
            "s14",
            dict(
                v1="1.4.6-SNAPSHOT (private-0986edd9-example)", op="__lt__", v2="1.4.6"
            ),
        ),
        (
            "s15",
            dict(
                v1="1.4.6-SNAPSHOT (private-0986edd9-example)", op="__gt__", v2="1.4.5"
            ),
        ),
    ]

    def setUp(self):
        super(ModuleRegistryPluginInfoTestsWithScenarios, self).setUp()

        jjb_config = JJBConfig()
        jjb_config.validate()

        plugin_info = [
            {"shortName": "HerpDerpPlugin", "longName": "Blah Blah Blah Plugin"}
        ]
        plugin_info.append(
            {
                "shortName": "JankyPlugin1",
                "longName": "Not A Real Plugin",
                "version": self.v1,
            }
        )

        self.addDetail("plugin_info", text_content(str(plugin_info)))
        self.registry = ModuleRegistry(jjb_config, plugin_info)

    def tearDown(self):
        super(ModuleRegistryPluginInfoTestsWithScenarios, self).tearDown()

    def test_get_plugin_info_dict(self):
        """
        The goal of this test is to validate that the plugin_info returned by
        ModuleRegistry.get_plugin_info is a dictionary whose key 'shortName' is
        the same value as the string argument passed to
        ModuleRegistry.get_plugin_info.
        """
        plugin_name = "JankyPlugin1"
        plugin_info = self.registry.get_plugin_info(plugin_name)

        self.assertIsInstance(plugin_info, dict)
        self.assertEqual(plugin_info["shortName"], plugin_name)

    def test_get_plugin_info_dict_using_longName(self):
        """
        The goal of this test is to validate that the plugin_info returned by
        ModuleRegistry.get_plugin_info is a dictionary whose key 'longName' is
        the same value as the string argument passed to
        ModuleRegistry.get_plugin_info.
        """
        plugin_name = "Blah Blah Blah Plugin"
        plugin_info = self.registry.get_plugin_info(plugin_name)

        self.assertIsInstance(plugin_info, dict)
        self.assertEqual(plugin_info["longName"], plugin_name)

    def test_get_plugin_info_dict_no_plugin(self):
        """
        The goal of this test case is to validate the behavior of
        ModuleRegistry.get_plugin_info when the given plugin cannot be found in
        ModuleRegistry's internal representation of the plugins_info.
        """
        plugin_name = "PluginDoesNotExist"
        plugin_info = self.registry.get_plugin_info(plugin_name)

        self.assertIsInstance(plugin_info, dict)
        self.assertEqual(plugin_info, {})

    def test_get_plugin_info_dict_no_version(self):
        """
        The goal of this test case is to validate the behavior of
        ModuleRegistry.get_plugin_info when the given plugin shortName returns
        plugin_info dict that has no version string. In a sane world where
        plugin frameworks like Jenkins' are sane this should never happen, but
        I am including this test and the corresponding default behavior
        because, well, it's Jenkins.
        """
        plugin_name = "HerpDerpPlugin"
        plugin_info = self.registry.get_plugin_info(plugin_name)

        self.assertIsInstance(plugin_info, dict)
        self.assertEqual(plugin_info["shortName"], plugin_name)
        self.assertEqual(plugin_info["version"], "0")

    def test_plugin_version_comparison(self):
        """
        The goal of this test case is to validate that valid tuple versions are
        ordinally correct. That is, for each given scenario, v1.op(v2)==True
        where 'op' is the equality operator defined for the scenario.
        """
        plugin_name = "JankyPlugin1"
        plugin_info = self.registry.get_plugin_info(plugin_name)
        v1 = plugin_info.get("version")

        op = getattr(pkg_resources.parse_version(v1), self.op)
        test = op(pkg_resources.parse_version(self.v2))

        self.assertTrue(
            test,
            msg="Unexpectedly found {0} {2} {1} == False "
            "when comparing versions!".format(v1, self.v2, self.op),
        )



#!/usr/bin/env python
# Copyright (C) 2015 OpenStack, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Manage Jenkins plugin module registry.

import logging
import operator
import pkg_resources
import re
import types

from jenkins_jobs.errors import JenkinsJobsException
from jenkins_jobs.formatter import deep_format
from jenkins_jobs.local_yaml import Jinja2Loader

__all__ = ["ModuleRegistry"]

logger = logging.getLogger(__name__)


class ModuleRegistry(object):
    _entry_points_cache = {}
    _component_type_cache = {}

    def __init__(self, jjb_config, plugins_list=None):
        self.modules = []
        self.modules_by_component_type = {}
        self.handlers = {}
        self.jjb_config = jjb_config
        self.masked_warned = {}

        if plugins_list is None:
            self.plugins_dict = {}
        else:
            self.plugins_dict = self._get_plugins_info_dict(plugins_list)

        for entrypoint in pkg_resources.iter_entry_points(group="jenkins_jobs.modules"):
            Mod = entrypoint.load()
            mod = Mod(self)
            self.modules.append(mod)
            self.modules.sort(key=operator.attrgetter("sequence"))
            if mod.component_type is not None:
                self.modules_by_component_type[mod.component_type] = entrypoint

    @staticmethod
    def _get_plugins_info_dict(plugins_list):
        def mutate_plugin_info(plugin_info):
            """
            We perform mutations on a single member of plugin_info here, then
            return a dictionary with the longName and shortName of the plugin
            mapped to its plugin info dictionary.
            """
            version = plugin_info.get("version", "0")
            plugin_info["version"] = re.sub(
                r"(.*)-(?:SNAPSHOT|BETA).*", r"\g<1>.preview", version
            )

            aliases = []
            for key in ["longName", "shortName"]:
                value = plugin_info.get(key, None)
                if value is not None:
                    aliases.append(value)

            plugin_info_dict = {}
            for name in aliases:
                plugin_info_dict[name] = plugin_info

            return plugin_info_dict

        list_of_dicts = [mutate_plugin_info(v) for v in plugins_list]

        plugins_info_dict = {}
        for d in list_of_dicts:
            plugins_info_dict.update(d)

        return plugins_info_dict

    def get_plugin_info(self, plugin_name):
        """Provide information about plugins within a module's impl of Base.gen_xml.

        The return value is a dictionary with data obtained directly from a
        running Jenkins instance.
        This allows module authors to differentiate generated XML output based
        on information such as specific plugin versions.

        :arg str plugin_name: Either the shortName or longName of a plugin
          as see in a query that looks like:
          ``http://<jenkins-hostname>/pluginManager/api/json?pretty&depth=2``

        During a 'test' run, it is possible to override JJB's query to a live
        Jenkins instance by passing it a path to a file containing a YAML list
        of dictionaries that mimics the plugin properties you want your test
        output to reflect::

          jenkins-jobs test -p /path/to/plugins-info.yaml

        Below is example YAML that might be included in
        /path/to/plugins-info.yaml.

        .. literalinclude:: /../../tests/cmd/fixtures/plugins-info.yaml

        """
        return self.plugins_dict.get(plugin_name, {})

    def registerHandler(self, category, name, method):
        cat_dict = self.handlers.get(category, {})
        if not cat_dict:
            self.handlers[category] = cat_dict
        cat_dict[name] = method

    def getHandler(self, category, name):
        return self.handlers[category][name]

    @property
    def parser_data(self):
        return self.__parser_data

    def set_parser_data(self, parser_data):
        self.__parser_data = parser_data

    def get_component_list_type(self, entry_point):
        if entry_point in self._component_type_cache:
            return self._component_type_cache[entry_point]

        # pkg_resources.EntryPoint.load() is costly, cache it.
        component_list_type = entry_point.load().component_list_type
        logging.info("Caching type %s of %s", component_list_type, entry_point)
        self._component_type_cache[entry_point] = component_list_type

        return component_list_type

    def dispatch(self, component_type, xml_parent, component, template_data={}):
        """This is a method that you can call from your implementation of
        Base.gen_xml or component.  It allows modules to define a type
        of component, and benefit from extensibility via Python
        entry points and Jenkins Job Builder :ref:`Macros <macro>`.

        :arg str component_type: the name of the component
          (e.g., `builder`)
        :arg YAMLParser parser: the global YAML Parser
        :arg Element xml_parent: the parent XML element
        :arg dict template_data: values that should be interpolated into
          the component definition

        See :py:class:`jenkins_jobs.modules.base.Base` for how to register
        components of a module.

        See the Publishers module for a simple example of how to use
        this method.
        """

        if component_type not in self.modules_by_component_type:
            raise JenkinsJobsException(
                "Unknown component type: " "'{0}'.".format(component_type)
            )

        entry_point = self.modules_by_component_type[component_type]
        component_list_type = self.get_component_list_type(entry_point)

        if isinstance(component, dict):
            # The component is a singleton dictionary of name: dict(args)
            name, component_data = next(iter(component.items()))
            if template_data or isinstance(component_data, Jinja2Loader):
                # Template data contains values that should be interpolated
                # into the component definition.  To handle Jinja2 templates
                # that don't contain any variables, we also deep format those.
                try:
                    component_data = deep_format(
                        component_data,
                        template_data,
                        self.jjb_config.yamlparser["allow_empty_variables"],
                    )
                except Exception:
                    logging.error(
                        "Failure formatting component ('%s') data '%s'",
                        name,
                        component_data,
                    )
                    raise
        else:
            # The component is a simple string name, eg "run-tests"
            name = component
            component_data = {}

        # Look for a component function defined in an entry point
        eps = self._entry_points_cache.get(component_list_type)
        if eps is None:
            logging.debug("Caching entrypoints for %s" % component_list_type)
            module_eps = []
            # auto build entry points by inferring from base component_types
            mod = pkg_resources.EntryPoint(
                "__all__", entry_point.module_name, dist=entry_point.dist
            )

            Mod = mod.load()
            func_eps = [
                Mod.__dict__.get(a)
                for a in dir(Mod)
                if isinstance(Mod.__dict__.get(a), types.FunctionType)
            ]
            for func_ep in func_eps:
                try:
                    # extract entry point based on docstring
                    name_line = func_ep.__doc__.split("\n")
                    if not name_line[0].startswith("yaml:"):
                        logger.debug("Ignoring '%s' as an entry point" % name_line)
                        continue
                    ep_name = name_line[0].split(" ")[1]
                except (AttributeError, IndexError):
                    # AttributeError by docstring not being defined as
                    # a string to have split called on it.
                    # IndexError raised by name_line not containing anything
                    # after the 'yaml:' string.
                    logger.debug(
                        "Not including func '%s' as an entry point" % func_ep.__name__
                    )
                    continue

                module_eps.append(
                    pkg_resources.EntryPoint(
                        ep_name,
                        entry_point.module_name,
                        dist=entry_point.dist,
                        attrs=(func_ep.__name__,),
                    )
                )
                logger.debug(
                    "Adding auto EP '%s=%s:%s'"
                    % (ep_name, entry_point.module_name, func_ep.__name__)
                )

            # load from explicitly defined entry points
            module_eps.extend(
                list(
                    pkg_resources.iter_entry_points(
                        group="jenkins_jobs.{0}".format(component_list_type)
                    )
                )
            )

            eps = {}
            for module_ep in module_eps:
                if module_ep.name in eps:
                    raise JenkinsJobsException(
                        "Duplicate entry point found for component type: "
                        "'{0}', '{0}',"
                        "name: '{1}'".format(component_type, name)
                    )

                eps[module_ep.name] = module_ep.load()

            # cache both sets of entry points
            self._entry_points_cache[component_list_type] = eps
            logger.debug("Cached entry point group %s = %s", component_list_type, eps)

        # check for macro first
        component = self.parser_data.get(component_type, {}).get(name)
        if component:
            if name in eps and name not in self.masked_warned:
                self.masked_warned[name] = True
                logger.warning(
                    "You have a macro ('%s') defined for '%s' "
                    "component type that is masking an inbuilt "
                    "definition" % (name, component_type)
                )

            for b in component[component_list_type]:
                # Pass component_data in as template data to this function
                # so that if the macro is invoked with arguments,
                # the arguments are interpolated into the real defn.
                self.dispatch(component_type, xml_parent, b, component_data)
        elif name in eps:
            func = eps[name]
            func(self, xml_parent, component_data)
        else:
            raise JenkinsJobsException(
                "Unknown entry point or macro '{0}' "
                "for component type: '{1}'.".format(name, component_type)
            )



#!/usr/bin/env python
# Copyright (C) 2013 Hewlett-Packard.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Provides local yaml parsing classes and extend yaml module

"""Custom application specific yamls tags are supported to provide
enhancements when reading yaml configuration.

Action Tags
^^^^^^^^^^^

These allow manipulation of data being stored in one layout in the source
yaml for convenience and/or clarity, to another format to be processed by
the targeted module instead of requiring all modules in JJB being capable
of supporting multiple input formats.

The tag ``!join:`` will treat the first element of the following list as
the delimiter to use, when joining the remaining elements into a string
and returning a single string to be consumed by the specified module option.

This allows users to maintain elements of data in a list structure for ease
of review/maintenance, and have the yaml parser convert it to a string for
consumption as any argument for modules. The main expected use case is to
allow for generic plugin data such as shell properties to be populated from
a list construct which the yaml parser converts to a single string, instead
of trying to support this within the module code which would require a
templating engine similar to Jinja.

Generic Example:

    .. literalinclude:: /../../tests/localyaml/fixtures/joinlists.yaml


Environment Inject:

    .. literalinclude:: /../../tests/yamlparser/fixtures/string_join.yaml


While this mechanism can also be used items where delimiters are supported by
the module, that should be considered a bug that the existing code doesn't
handle being provided a list and delimiter to perform the correct conversion
for you. Should you discover a module that takes arguments with delimiters and
the existing JJB codebase does not handle accepting lists, then this can be
used as a temporary solution in place of using very long strings:

Extended Params Example:

    .. literalinclude::
        /../../tests/parameters/fixtures/extended-choice-param-full.yaml


Inclusion Tags
^^^^^^^^^^^^^^

These allow inclusion of arbitrary files as a method of having blocks of data
managed separately to the yaml job configurations. A specific usage of this is
inlining scripts contained in separate files, although such tags may also be
used to simplify usage of macros or job templates.

The tag ``!include:`` will treat the following string as file which should be
parsed as yaml configuration data.

Example:

    .. literalinclude:: /../../tests/localyaml/fixtures/include001.yaml

    contents of include001.yaml.inc:

    .. literalinclude:: /../../tests/yamlparser/fixtures/include001.yaml.inc


The tag ``!include-raw:`` will treat the given string or list of strings as
filenames to be opened as one or more data blob, which should be read into
the calling yaml construct without any further parsing. Any data in a file
included through this tag, will be treated as string data.

Examples:

    .. literalinclude:: /../../tests/localyaml/fixtures/include-raw001.yaml

    contents of include-raw001-hello-world.sh:

        .. literalinclude::
            /../../tests/localyaml/fixtures/include-raw001-hello-world.sh

    contents of include-raw001-vars.sh:

        .. literalinclude::
            /../../tests/localyaml/fixtures/include-raw001-vars.sh

    using a list of files:

    .. literalinclude::
        /../../tests/localyaml/fixtures/include-raw-multi001.yaml

The tag ``!include-raw-escape:`` treats the given string or list of strings as
filenames to be opened as one or more data blobs, which should be escaped
before being read in as string data. This allows job-templates to use this tag
to include scripts from files without needing to escape braces in the original
file.

.. warning::

    When used as a macro ``!include-raw-escape:`` should only be used if
    parameters are passed into the escaped file and you would like to escape
    those parameters. If the file does not have any jjb parameters passed into
    it then ``!include-raw:`` should be used instead otherwise you will run
    into an interesting issue where ``include-raw-escape:`` actually adds
    additional curly braces around existing curly braces. For example
    ${PROJECT} becomes ${{PROJECT}} which may break bash scripts.

Examples:

    .. literalinclude::
        /../../tests/localyaml/fixtures/include-raw-escaped001.yaml

    contents of include-raw001-hello-world.sh:

        .. literalinclude::
            /../../tests/localyaml/fixtures/include-raw001-hello-world.sh

    contents of include-raw001-vars.sh:

        .. literalinclude::
            /../../tests/localyaml/fixtures/include-raw001-vars.sh

    using a list of files:

    .. literalinclude::
        /../../tests/localyaml/fixtures/include-raw-escaped-multi001.yaml


For all the multi file includes, the files are simply appended using a newline
character.


To allow for job templates to perform substitution on the path names, when a
filename containing a python format placeholder is encountered, lazy loading
support is enabled, where instead of returning the contents back during yaml
parsing, it is delayed until the variable substitution is performed.

Example:

    .. literalinclude:: /../../tests/yamlparser/fixtures/lazy-load-jobs001.yaml

    using a list of files:

    .. literalinclude::
        /../../tests/yamlparser/fixtures/lazy-load-jobs-multi001.yaml

.. note::

    Because lazy-loading involves performing the substitution on the file
    name, it means that jenkins-job-builder can not call the variable
    substitution on the contents of the file. This means that the
    ``!include-raw:`` tag will behave as though ``!include-raw-escape:`` tag
    was used instead whenever name substitution on the filename is to be
    performed.

    Given the behaviour described above, when substitution is to be performed
    on any filename passed via ``!include-raw-escape:`` the tag will be
    automatically converted to ``!include-raw:`` and no escaping will be
    performed.


The tag ``!include-jinja2:`` will treat the given string or list of strings as
filenames to be opened as Jinja2 templates, which should be rendered to a
string and included in the calling YAML construct.  (This is analogous to the
templating that will happen with ``!include-raw``.)

Examples:

    .. literalinclude:: /../../tests/yamlparser/fixtures/jinja01.yaml

    contents of jinja01.yaml.inc:

        .. literalinclude:: /../../tests/yamlparser/fixtures/jinja01.yaml.inc


The tag ``!j2:`` takes a string and treats it as a Jinja2 template.  It will be
rendered (with the variables in that context) and included in the calling YAML
construct.

Examples:

    .. literalinclude:: /../../tests/yamlparser/fixtures/jinja-string01.yaml

The tag ``!j2-yaml:`` is similar to the ``!j2:`` tag, just that it loads the
Jinja-rendered string as YAML and embeds it in the calling YAML construct. This
provides a very flexible and convenient way of generating pieces of YAML
structures. One of use cases is defining complex YAML structures with much
simpler configuration, without any duplication.

Examples:

    .. literalinclude:: /../../tests/yamlparser/fixtures/jinja-yaml01.yaml

Another use case is controlling lists dynamically, like conditionally adding
list elements based on project configuration.

Examples:

    .. literalinclude:: /../../tests/yamlparser/fixtures/jinja-yaml02.yaml

"""

import functools
import io
import logging
import os
import re
import copy

import jinja2
import yaml
from yaml.constructor import BaseConstructor
from yaml.representer import BaseRepresenter
from yaml import YAMLObject

from collections import OrderedDict


logger = logging.getLogger(__name__)


class OrderedConstructor(BaseConstructor):
    """The default constructor class for PyYAML loading uses standard python
    dictionaries which can have randomized ordering enabled (default in
    CPython from version 3.3). The order of the XML elements being outputted
    is both important for tests and for ensuring predictable generation based
    on the source. This subclass overrides this behaviour to ensure that all
    dict's created make use of OrderedDict to have iteration of keys to always
    follow the order in which the keys were inserted/created.
    """

    def construct_yaml_map(self, node):
        data = OrderedDict()
        yield data
        value = self.construct_mapping(node)

        if isinstance(node, yaml.MappingNode):
            self.flatten_mapping(node)
        else:
            raise yaml.constructor.ConstructorError(
                None,
                None,
                "expected a mapping node, but found %s" % node.id,
                node.start_mark,
            )

        mapping = OrderedDict()
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=False)
            try:
                hash(key)
            except TypeError as exc:
                raise yaml.constructor.ConstructorError(
                    "while constructing a mapping",
                    node.start_mark,
                    "found unacceptable key (%s)" % exc,
                    key_node.start_mark,
                )
            value = self.construct_object(value_node, deep=False)
            mapping[key] = value
        data.update(mapping)


class OrderedRepresenter(BaseRepresenter):
    def represent_yaml_mapping(self, mapping, flow_style=None):
        tag = u"tag:yaml.org,2002:map"
        node = self.represent_mapping(tag, mapping, flow_style=flow_style)
        return node


class LocalAnchorLoader(yaml.Loader):
    """Subclass for yaml.Loader which keeps Alias between calls"""

    anchors = {}

    def __init__(self, *args, **kwargs):
        super(LocalAnchorLoader, self).__init__(*args, **kwargs)
        self.anchors = LocalAnchorLoader.anchors

    @classmethod
    def reset_anchors(cls):
        cls.anchors = {}

    # override the default composer to skip resetting the anchors at the
    # end of the current document
    def compose_document(self):
        # Drop the DOCUMENT-START event.
        self.get_event()
        # Compose the root node.
        node = self.compose_node(None, None)
        # Drop the DOCUMENT-END event.
        self.get_event()
        return node


class LocalLoader(OrderedConstructor, LocalAnchorLoader):
    """Subclass for yaml.Loader which handles storing the search_path and
    escape_callback functions for use by the custom YAML objects to find files
    and escape the content where required.

    Constructor access a list of search paths to look under for the given
    file following each tag, taking the first match found. Search path by
    default will include the same directory as the yaml file and the current
    working directory.


    Loading::

        # use the load function provided in this module
        import local_yaml
        data = local_yaml.load(io.open(fn, 'r', encoding='utf-8'))


        # Loading by providing the alternate class to the default yaml load
        from local_yaml import LocalLoader
        data = yaml.load(io.open(fn, 'r', encoding='utf-8'), LocalLoader)

        # Loading with a search path
        from local_yaml import LocalLoader
        import functools
        data = yaml.load(io.open(fn, 'r', encoding='utf-8'),
                         functools.partial(LocalLoader, search_path=['path']))

    """

    def __init__(self, *args, **kwargs):
        # make sure to pop off any local settings before passing to
        # the parent constructor as any unknown args may cause errors.
        self.search_path = list()
        if "search_path" in kwargs:
            for p in kwargs.pop("search_path"):
                logger.debug("Adding '{0}' to search path for include tags".format(p))
                self.search_path.append(os.path.normpath(p))

        if "escape_callback" in kwargs:
            self.escape_callback = kwargs.pop("escape_callback")
        else:
            self.escape_callback = self._escape

        super(LocalLoader, self).__init__(*args, **kwargs)

        # constructor to preserve order of maps and ensure that the order of
        # keys returned is consistent across multiple python versions
        self.add_constructor(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            type(self).construct_yaml_map,
        )

        if hasattr(self.stream, "name"):
            self.search_path.append(os.path.normpath(os.path.dirname(self.stream.name)))
        self.search_path.append(os.path.normpath(os.path.curdir))

    def _escape(self, data):
        return re.sub(r"({|})", r"\1\1", data)


class LocalDumper(OrderedRepresenter, yaml.Dumper):
    def __init__(self, *args, **kwargs):
        super(LocalDumper, self).__init__(*args, **kwargs)

        # representer to ensure conversion back looks like normal
        # mapping and hides that we use OrderedDict internally
        self.add_representer(OrderedDict, type(self).represent_yaml_mapping)
        # convert any tuples to lists as the JJB input is generally
        # in list format
        self.add_representer(tuple, type(self).represent_list)


class BaseYAMLObject(YAMLObject):
    yaml_loader = LocalLoader
    yaml_dumper = LocalDumper


class J2Yaml(BaseYAMLObject):
    yaml_tag = u"!j2-yaml:"

    @classmethod
    def from_yaml(cls, loader, node):
        return Jinja2YamlLoader(node.value, loader.search_path)


class J2String(BaseYAMLObject):
    yaml_tag = u"!j2:"

    @classmethod
    def from_yaml(cls, loader, node):
        return Jinja2Loader(node.value, loader.search_path)


class YamlListJoin(BaseYAMLObject):
    yaml_tag = u"!join:"

    @classmethod
    def from_yaml(cls, loader, node):
        if isinstance(node, yaml.SequenceNode):
            delimiter = node.value[0].value
            if not isinstance(node.value[1], yaml.SequenceNode):
                raise yaml.constructor.ConstructorError(
                    None,
                    None,
                    "expected sequence node for join data, but "
                    "found %s" % node.value[1].id,
                    node.start_mark,
                )

            return delimiter.join((v.value for v in node.value[1].value))
        else:
            raise yaml.constructor.ConstructorError(
                None,
                None,
                "expected sequence node, but found %s" % node.id,
                node.start_mark,
            )


class YamlInclude(BaseYAMLObject):
    yaml_tag = u"!include:"

    @classmethod
    def _find_file(cls, filename, search_path):
        for dirname in search_path:
            candidate = os.path.expanduser(os.path.join(dirname, filename))
            if os.path.isfile(candidate):
                logger.debug(
                    "Including file '{0}' from path '{1}'".format(filename, dirname)
                )
                return candidate
        return filename

    @classmethod
    def _open_file(cls, loader, node):
        node_str = loader.construct_yaml_str(node)
        try:
            node_str.format()
        except KeyError:
            return cls._lazy_load(loader, cls.yaml_tag, node)

        filename = cls._find_file(node_str, loader.search_path)
        try:
            with io.open(filename, "r", encoding="utf-8") as f:
                return f.read()
        except Exception:
            logger.error(
                "Failed to include file using search path: '{0}'".format(
                    ":".join(loader.search_path)
                )
            )
            raise

    @classmethod
    def _from_file(cls, loader, node):
        contents = cls._open_file(loader, node)
        if isinstance(contents, LazyLoader):
            return contents

        data = yaml.load(
            contents, functools.partial(cls.yaml_loader, search_path=loader.search_path)
        )
        return data

    @classmethod
    def _lazy_load(cls, loader, tag, node_str):
        logger.info("Lazy loading of file template '{0}' enabled".format(node_str))
        return LazyLoader((cls, loader, node_str))

    @classmethod
    def from_yaml(cls, loader, node):
        if isinstance(node, yaml.ScalarNode):
            return cls._from_file(loader, node)
        elif isinstance(node, yaml.SequenceNode):
            contents = [
                cls._from_file(loader, scalar_node) for scalar_node in node.value
            ]
            if any(isinstance(s, CustomLoader) for s in contents):
                return CustomLoaderCollection(contents)

            return u"\n".join(contents)
        else:
            raise yaml.constructor.ConstructorError(
                None,
                None,
                "expected either a sequence or scalar node, but " "found %s" % node.id,
                node.start_mark,
            )


class YamlIncludeRaw(YamlInclude):
    yaml_tag = u"!include-raw:"

    @classmethod
    def _from_file(cls, loader, node):
        return cls._open_file(loader, node)


class YamlIncludeRawEscape(YamlIncludeRaw):
    yaml_tag = u"!include-raw-escape:"

    @classmethod
    def from_yaml(cls, loader, node):
        data = YamlIncludeRaw.from_yaml(loader, node)
        if isinstance(data, LazyLoader):
            logger.warning(
                "Replacing %s tag with %s since lazy loading means "
                "file contents will not be deep formatted for "
                "variable substitution.",
                cls.yaml_tag,
                YamlIncludeRaw.yaml_tag,
            )
            return data
        else:
            return loader.escape_callback(data)


class YamlIncludeJinja2(YamlIncludeRaw):
    yaml_tag = u"!include-jinja2:"

    @classmethod
    def _from_file(cls, loader, node):
        contents = cls._open_file(loader, node)
        if isinstance(contents, LazyLoader):
            return contents
        return Jinja2Loader(contents, loader.search_path)


class DeprecatedTag(BaseYAMLObject):
    @classmethod
    def from_yaml(cls, loader, node):
        logger.warning(
            "tag '%s' is deprecated, switch to using '%s'",
            cls.yaml_tag,
            cls._new.yaml_tag,
        )
        return cls._new.from_yaml(loader, node)


class YamlIncludeDeprecated(DeprecatedTag):
    yaml_tag = u"!include"
    _new = YamlInclude


class YamlIncludeRawDeprecated(DeprecatedTag):
    yaml_tag = u"!include-raw"
    _new = YamlIncludeRaw


class YamlIncludeRawEscapeDeprecated(DeprecatedTag):
    yaml_tag = u"!include-raw-escape"
    _new = YamlIncludeRawEscape


class CustomLoader(object):
    """Parent class for non-standard loaders."""


class Jinja2Loader(CustomLoader):
    """A loader for Jinja2-templated files."""

    def __init__(self, contents, search_path):
        # capture template contents and search paths on loader creation.
        self._contents = contents
        self._search_path = search_path
        self._template = None
        self._loader = None

    def __deepcopy__(self, memo):
        # Jinja 2 templates are not deepcopy-able so just pass around
        # the search_path and contents.
        return Jinja2Loader(self._contents, self._search_path)

    def format(self, **kwargs):
        # Wait until first render call to create a template then save
        # the template on this instance for faster rendering.
        if not self._template:
            self._template = jinja2.Template(self._contents)
            self._template.environment.undefined = jinja2.StrictUndefined
            self._template.environment.loader = jinja2.FileSystemLoader(
                self._search_path
            )
            # Preserve this loader if it hasn't been overwritten
            # elsewhere.
            self._loader = self._template.environment.loader
        self._template.environment.loader = self._loader
        return self._template.render(kwargs)

    def get_object_to_format(self):
        return self


class LateYamlLoader(CustomLoader):
    """A loader for data rendered via Jinja2, to be loaded as YAML and then deep formatted."""

    def __init__(self, yaml_str, loader):
        self._yaml_str = yaml_str
        self._loader = loader

    def get_object_to_format(self):
        return load(self._yaml_str, search_path=self._loader._search_path)


class Jinja2YamlLoader(Jinja2Loader):
    def format(self, **kwargs):
        yaml_str = super(Jinja2YamlLoader, self).format(**kwargs)
        return LateYamlLoader(yaml_str, self)


class CustomLoaderCollection(object):
    """Helper class to format a collection of CustomLoader objects"""

    def __init__(self, sequence):
        self._data = sequence

    def format(self, *args, **kwargs):
        return u"\n".join(item.format(*args, **kwargs) for item in self._data)


class LazyLoader(CustomLoader):
    """Helper class to provide lazy loading of files included using !include*
    tags where the path to the given file contains unresolved placeholders.
    """

    def __init__(self, data):
        # str subclasses can only have one argument, so assume it is a tuple
        # being passed and unpack as needed
        self._cls, self._loader, self._node = data

    def __str__(self):
        return "%s %s" % (self._cls.yaml_tag, self._node.value)

    def __repr__(self):
        return "%s %s" % (self._cls.yaml_tag, self._node.value)

    def __deepcopy__(self, memodict={}):
        return LazyLoader(
            (copy.deepcopy(self._cls), self._loader, copy.deepcopy(self._node))
        )

    def format(self, *args, **kwargs):
        node = yaml.ScalarNode(
            tag=self._node.tag, value=self._node.value.format(*args, **kwargs)
        )
        return self._cls.from_yaml(self._loader, node)


def load(stream, retain_anchors=False, **kwargs):
    if not retain_anchors:
        LocalAnchorLoader.reset_anchors()
    return yaml.load(stream, functools.partial(LocalLoader, **kwargs))


def dump(data, stream=None, **kwargs):
    return yaml.dump(data, stream, Dumper=LocalDumper, **kwargs)



import io
import os

from tests.base import mock
from tests.cmd.test_cmd import CmdTestsBase

from jenkins_jobs.cli import entry
from jenkins_jobs import builder

patch = mock.patch


@mock.patch("jenkins_jobs.builder.JenkinsManager.get_plugins_info", mock.MagicMock)
class TestConfigs(CmdTestsBase):

    global_conf = "/etc/jenkins_jobs/jenkins_jobs.ini"
    user_conf = os.path.join(
        os.path.expanduser("~"), ".config", "jenkins_jobs", "jenkins_jobs.ini"
    )
    local_conf = os.path.join(os.path.dirname(__file__), "jenkins_jobs.ini")

    def test_use_global_config(self):
        """
        Verify that JJB uses the global config file by default
        """

        args = ["test", "foo"]
        conffp = io.open(self.default_config_file, "r", encoding="utf-8")

        with patch("os.path.isfile", return_value=True) as m_isfile:

            def side_effect(path):
                if path == self.global_conf:
                    return True
                return False

            m_isfile.side_effect = side_effect

            with patch("io.open", return_value=conffp) as m_open:
                entry.JenkinsJobs(args, config_file_required=True)
                m_open.assert_called_with(self.global_conf, "r", encoding="utf-8")

    def test_use_config_in_user_home(self):
        """
        Verify that JJB uses config file in user home folder
        """

        args = ["test", "foo"]

        conffp = io.open(self.default_config_file, "r", encoding="utf-8")
        with patch("os.path.isfile", return_value=True) as m_isfile:

            def side_effect(path):
                if path == self.user_conf:
                    return True
                return False

            m_isfile.side_effect = side_effect
            with patch("io.open", return_value=conffp) as m_open:
                entry.JenkinsJobs(args, config_file_required=True)
                m_open.assert_called_with(self.user_conf, "r", encoding="utf-8")

    def test_non_existing_config_dir(self):
        """
        Run test mode and pass a non-existing configuration directory
        """
        args = ["--conf", self.default_config_file, "test", "foo"]
        jenkins_jobs = entry.JenkinsJobs(args)
        self.assertRaises(IOError, jenkins_jobs.execute)

    def test_non_existing_config_file(self):
        """
        Run test mode and pass a non-existing configuration file
        """
        args = ["--conf", self.default_config_file, "test", "non-existing.yaml"]
        jenkins_jobs = entry.JenkinsJobs(args)
        self.assertRaises(IOError, jenkins_jobs.execute)

    def test_config_options_not_replaced_by_cli_defaults(self):
        """
        Run test mode and check config settings from conf file retained
        when none of the global CLI options are set.
        """
        config_file = os.path.join(self.fixtures_path, "settings_from_config.ini")
        args = ["--conf", config_file, "test", "dummy.yaml"]
        jenkins_jobs = entry.JenkinsJobs(args)
        jjb_config = jenkins_jobs.jjb_config
        self.assertEqual(jjb_config.jenkins["user"], "jenkins_user")
        self.assertEqual(jjb_config.jenkins["password"], "jenkins_password")
        self.assertEqual(jjb_config.builder["ignore_cache"], True)
        self.assertEqual(jjb_config.builder["flush_cache"], True)
        self.assertEqual(jjb_config.builder["update"], "all")
        self.assertEqual(jjb_config.yamlparser["allow_empty_variables"], True)

    def test_config_options_overriden_by_cli(self):
        """
        Run test mode and check config settings from conf file retained
        when none of the global CLI options are set.
        """
        args = [
            "--user",
            "myuser",
            "--password",
            "mypassword",
            "--ignore-cache",
            "--flush-cache",
            "--allow-empty-variables",
            "test",
            "dummy.yaml",
        ]
        jenkins_jobs = entry.JenkinsJobs(args)
        jjb_config = jenkins_jobs.jjb_config
        self.assertEqual(jjb_config.jenkins["user"], "myuser")
        self.assertEqual(jjb_config.jenkins["password"], "mypassword")
        self.assertEqual(jjb_config.builder["ignore_cache"], True)
        self.assertEqual(jjb_config.builder["flush_cache"], True)
        self.assertEqual(jjb_config.yamlparser["allow_empty_variables"], True)

    @mock.patch("jenkins_jobs.cli.subcommand.update.JenkinsManager")
    def test_update_timeout_not_set(self, jenkins_mock):
        """Check that timeout is left unset

        Test that the Jenkins object has the timeout set on it only when
        provided via the config option.
        """

        path = os.path.join(self.fixtures_path, "cmd-002.yaml")
        args = ["--conf", self.default_config_file, "update", path]

        jenkins_mock.return_value.update_jobs.return_value = ([], 0)
        jenkins_mock.return_value.update_views.return_value = ([], 0)
        self.execute_jenkins_jobs_with_args(args)

        # validate that the JJBConfig used to initialize builder.Jenkins
        # contains the expected timeout value.

        jjb_config = jenkins_mock.call_args[0][0]
        self.assertEqual(jjb_config.jenkins["timeout"], builder._DEFAULT_TIMEOUT)

    @mock.patch("jenkins_jobs.cli.subcommand.update.JenkinsManager")
    def test_update_timeout_set(self, jenkins_mock):
        """Check that timeout is set correctly

        Test that the Jenkins object has the timeout set on it only when
        provided via the config option.
        """

        path = os.path.join(self.fixtures_path, "cmd-002.yaml")
        config_file = os.path.join(self.fixtures_path, "non-default-timeout.ini")
        args = ["--conf", config_file, "update", path]

        jenkins_mock.return_value.update_jobs.return_value = ([], 0)
        jenkins_mock.return_value.update_views.return_value = ([], 0)
        self.execute_jenkins_jobs_with_args(args)

        # validate that the JJBConfig used to initialize builder.Jenkins
        # contains the expected timeout value.

        jjb_config = jenkins_mock.call_args[0][0]
        self.assertEqual(jjb_config.jenkins["timeout"], 0.2)



"""Exception classes for jenkins_jobs errors"""

import inspect


def is_sequence(arg):
    return not hasattr(arg, "strip") and (
        hasattr(arg, "__getitem__") or hasattr(arg, "__iter__")
    )


class JenkinsJobsException(Exception):
    pass


class ModuleError(JenkinsJobsException):
    def get_module_name(self):
        frame = inspect.currentframe()
        co_name = frame.f_code.co_name
        module_name = "<unresolved>"
        while frame and co_name != "run":
            # XML generation called via dispatch
            if co_name == "dispatch":
                data = frame.f_locals
                module_name = "%s.%s" % (data["component_type"], data["name"])
                break
            # XML generation done directly by class using gen_xml or root_xml
            if co_name == "gen_xml" or co_name == "root_xml":
                data = frame.f_locals["data"]
                module_name = next(iter(data.keys()))
                break
            frame = frame.f_back
            co_name = frame.f_code.co_name

        return module_name


class InvalidAttributeError(ModuleError):
    def __init__(self, attribute_name, value, valid_values=None):
        message = "'{0}' is an invalid value for attribute {1}.{2}".format(
            value, self.get_module_name(), attribute_name
        )

        if is_sequence(valid_values):
            message += "\nValid values include: {0}".format(
                ", ".join("'{0}'".format(value) for value in valid_values)
            )

        super(InvalidAttributeError, self).__init__(message)


class MissingAttributeError(ModuleError):
    def __init__(self, missing_attribute, module_name=None):
        module = module_name or self.get_module_name()
        if is_sequence(missing_attribute):
            message = "One of {0} must be present in '{1}'".format(
                ", ".join("'{0}'".format(value) for value in missing_attribute), module
            )
        else:
            message = "Missing {0} from an instance of '{1}'".format(
                missing_attribute, module
            )

        super(MissingAttributeError, self).__init__(message)


class AttributeConflictError(ModuleError):
    def __init__(self, attribute_name, attributes_in_conflict, module_name=None):
        module = module_name or self.get_module_name()
        message = "Attribute '{0}' can not be used together with {1} in {2}".format(
            attribute_name,
            ", ".join("'{0}'".format(value) for value in attributes_in_conflict),
            module,
        )

        super(AttributeConflictError, self).__init__(message)


class YAMLFormatError(JenkinsJobsException):
    pass


class JJBConfigException(JenkinsJobsException):
    pass
